from __future__ import annotations

import math

import numpy as np
from numba import jit

from pythermalcomfort.classes_input import PHSInputs
from pythermalcomfort.classes_return import PHS
from pythermalcomfort.utilities import (
    Models,
    Postures,
    _check_standard_compliance_array,
    met_to_w_m2,
    p_sat,
)


def phs(
    tdb: float | list[float],
    tr: float | list[float],
    v: float | list[float],
    rh: float | list[float],
    met: float | list[float],
    clo: float | list[float],
    posture: str | list[str],
    wme: float | np.ndarray | list[float] | list[int] = 0,
    round_output: bool = True,
    model: str = Models.iso_7933_2023.value,
    **kwargs,
) -> PHS:
    """Calculate the Predicted Heat Strain (PHS) index based in compliance
    with the ISO 7933:2004 [7933ISO2004]_ or 2023 Standard [7933ISO2023]_. The ISO 7933
    provides a method for the analytical evaluation and interpretation of the thermal
    stress experienced by a subject in a hot environment. It describes a method for
    predicting the sweat rate and the internal core temperature that the human
    body will develop in response to the working conditions.

    The PHS model can be used to predict the: heat by respiratory convection, heat flow
    by respiratory evaporation, steady state mean skin temperature, instantaneous value
    of skin temperature, heat accumulation associated with the metabolic rate, maximum
    evaporative heat flow at the skin surface, predicted sweat rate, predicted evaporative
    heat flow, and rectal temperature.

    Parameters
    ----------
    tdb : float or list of floats
        Dry bulb air temperature, [°C].
    tr : float or list of floats
        Mean radiant temperature, [°C].
    v : float or list of floats
        Air speed, [m/s].
    rh : float or list of floats
        Relative humidity, [%].
    met : float or list of floats
        Metabolic rate, [met].
    clo : float or list of floats
        Clothing insulation, [clo].
    posture: string or list of strings
        a string value presenting posture of person "sitting", "standing", or "crouching"
    wme : float or list of floats
        external work, [met] default 0
    round_output : bool, optional
        If True, rounds output value. If False, it does not round it. Defaults to True.
    model : str, optional
        Select the model you want to use to calculate the PHS. The default option is
        "7933-2023", and the other option is "7933-2004".

    Other Parameters
    ----------------
    limit_inputs : bool, optional
        If True, limits the input parameters to the standard's applicability limits. Defaults to True.

        .. note::
            By default, if the inputs are outside the standard applicability limits the
            function returns nan. If False returns values even if input values are
            outside the applicability limits of the model.

            The 7933 limits are 15 < tdb [°C] < 50, 0 < tr [°C] < 60,
            0 < vr [m/s] < 3, 1.7 < met [met] < 7.5, and 0.1 < clo [clo] < 1.

    i_mst : float, optional
        Static moisture permeability index, [dimensionless]. Defaults to 0.38.
    a_p : float, optional
        Fraction of the body surface covered by the reflective clothing, [dimensionless]. Defaults to 0.54.
    drink : int, optional
        1 if workers can drink freely, 0 otherwise. Defaults to 1.
    weight : float, optional
        Body weight, [kg]. Defaults to 75.
    height : float, optional
        Height, [m]. Defaults to 1.8.
    walk_sp : float, optional
        Walking speed, [m/s]. Defaults to 0.
    theta : float, optional
        Angle between walking direction and wind direction, [degrees]. Defaults to 0.
    acclimatized : int, optional
        100 if acclimatized subject, 0 otherwise. Defaults to 100.
    duration : int, optional
        Duration of the work sequence, [minutes]. Defaults to 480.
    f_r : float, optional
        Emissivity of the reflective clothing, [dimensionless]. Defaults to 0.97 in the 2004 standard and
        0.42 in the 2023 standard.
    t_sk : float, optional
        Mean skin temperature when worker starts working, [°C]. Defaults to 34.1.
    t_cr : float, optional
        Mean core temperature when worker starts working, [°C]. Defaults to 36.8.
    t_re : float, optional
        Mean rectal temperature when worker starts working, [°C]. If False in the 2004 standard,
        then t_re = t_cr, whereas in the 2023 standard t_re = 36.8 °C
    t_cr_eq : float, optional
        Mean core temperature as a function of met when worker starts working, [°C]. If False in the 2004
        standard, then t_cr_eq = t_cr, whereas in the 2023 standard t_cr_eq = 36.8 °C.
    sweat_rate : float, optional
        Initial sweat rate, [W/m2]. Defaults to 0.

    Returns
    -------
    PHS
        A dataclass containing the Predicted Heat Strain. See :py:class:`~pythermalcomfort.classes_return.PHS` for more details.
        To access the individual attributes, use the corresponding attribute of the returned `PHS` instance, e.g., `result.t_re`.

    Examples
    --------
    .. code-block:: python

        from pythermalcomfort.models import phs

        result = phs(
            tdb=40,
            tr=40,
            rh=33.85,
            v=0.3,
            met=2.5,
            clo=0.5,
            posture="standing",
            wme=0,
        )
        print(result.t_re)  # 37.5

        result = phs(
            tdb=[40, 45],
            tr=[40, 45],
            v=[0.3, 0.4],
            rh=[33.85, 40],
            met=[2.5, 2.6],
            clo=[0.5, 0.6],
            posture=["standing", "standing"],
            wme=[0, 0],
        )
        print(result.t_re)  # [37.5 43.4]

    """
    if model not in [Models.iso_7933_2004.value, Models.iso_7933_2023.value]:
        error_msg = (
            f"Model '{model}' is not supported. "
            f"Supported models are: {Models.iso_7933_2004.value} and {Models.iso_7933_2023.value}."
        )
        raise ValueError(error_msg)

    PHSInputs(
        tdb=tdb,
        tr=tr,
        v=v,
        rh=rh,
        met=met,
        clo=clo,
        round_output=round_output,
        wme=wme,
        posture=posture,
    )

    default_kwargs = {
        "i_mst": 0.38,
        "a_p": 0.54,
        "drink": 1,
        "weight": 75,
        "height": 1.8,
        "walk_sp": 0,
        "theta": 0,
        "acclimatized": 100,
        "duration": 480,
        "f_r": 0.97,
        "t_sk": 34.1,
        "t_cr": 36.8,
        "t_re": False,
        "t_cr_eq": False,
        "t_sk_t_cr_wg": 0.3,
        "sweat_rate": 0,
        "limit_inputs": True,
    }

    if model == Models.iso_7933_2023.value:
        # override changed default kwargs for 2023 standard
        overrides_2023 = {
            "f_r": 0.42,
            "t_re": 36.8,
            "t_cr_eq": 36.8,
        }
        default_kwargs.update(overrides_2023)

    kwargs = {**default_kwargs, **kwargs}

    tdb = np.array(tdb)
    tr = np.array(tr)
    v = np.array(v)
    rh = np.array(rh)
    met = np.array(met) * met_to_w_m2
    clo = np.array(clo)
    wme = np.array(wme) * met_to_w_m2
    posture = np.array(posture)

    i_mst = kwargs["i_mst"]
    a_p = kwargs["a_p"]
    drink = kwargs["drink"]
    weight = kwargs["weight"]
    height = kwargs["height"]
    walk_sp = kwargs["walk_sp"]
    theta = kwargs["theta"]
    acclimatized = kwargs["acclimatized"]
    duration = kwargs["duration"]
    f_r = kwargs["f_r"]
    t_sk = kwargs["t_sk"]
    t_cr = kwargs["t_cr"]
    t_re = kwargs["t_re"]
    t_cr_eq = kwargs["t_cr_eq"]
    t_sk_t_cr_wg = kwargs["t_sk_t_cr_wg"]
    sweat_rate = kwargs["sweat_rate"]
    limit_inputs = kwargs["limit_inputs"]

    if model == Models.iso_7933_2023.value:
        p_a = 0.6105 * np.exp(17.27 * tdb / (tdb + 237.3)) * rh / 100
    else:  # model == Models.iso_7933_2004.value:
        p_a = p_sat(tdb) / 1000 * rh / 100

    acclimatized = int(acclimatized)
    if acclimatized not in [0, 100]:
        raise ValueError("Acclimatized should be 0 or 100")

    if drink not in [0, 1]:
        raise ValueError("Drink should be 0 or 1")

    if weight <= 0 or weight > 1000:
        raise ValueError(
            "The weight of the person should be in kg and it cannot exceed 1000",
        )

    if not t_re:
        t_re = t_cr
    if not t_cr_eq:
        t_cr_eq = t_cr

    (
        t_re,
        t_sk,
        t_cr,
        t_cr_eq,
        t_sk_t_cr_wg,
        sweat_rate,
        sw_tot_g,
        d_lim_loss_50,
        d_lim_loss_95,
        d_lim_t_re,
    ) = _phs_optimized(
        tdb=tdb,
        tr=tr,
        v=v,
        p_a=p_a,
        met=met,
        clo=clo,
        posture=posture,
        drink=drink,
        acclimatized=acclimatized,
        weight=weight,
        wme=wme,
        i_mst=i_mst,
        a_p=a_p,
        height=height,
        walk_sp=walk_sp,
        theta=theta,
        duration=duration,
        f_r=f_r,
        t_sk=t_sk,
        t_cr=t_cr,
        t_re=t_re,
        t_cr_eq=t_cr_eq,
        t_sk_t_cr_wg=t_sk_t_cr_wg,
        sw_tot=sweat_rate,
        model=model,
    )

    output = {
        "t_re": t_re,
        "t_sk": t_sk,
        "t_cr": t_cr,
        "t_cr_eq": t_cr_eq,
        "t_sk_t_cr_wg": t_sk_t_cr_wg,
        "d_lim_loss_50": d_lim_loss_50,
        "d_lim_loss_95": d_lim_loss_95,
        "d_lim_t_re": d_lim_t_re,
        "water_loss_watt": sweat_rate,
        "water_loss": sw_tot_g,
    }

    if limit_inputs:
        (
            tdb_valid,
            tr_valid,
            v_valid,
            p_a_valid,
            met_valid,
            clo_valid,
        ) = _check_standard_compliance_array(
            model,
            tdb=tdb,
            tr=tr,
            v=v,
            met=met,
            clo=clo,
            p_a=p_a,
        )
        all_valid = ~(
            np.isnan(tdb_valid)
            | np.isnan(tr_valid)
            | np.isnan(v_valid)
            | np.isnan(p_a_valid)
            | np.isnan(met_valid)
            | np.isnan(clo_valid)
        )
        for key in output:
            output[key] = np.where(all_valid, output[key], np.nan)

    if round_output:
        for key in output:
            if key != "t_sk_t_cr_wg":
                output[key] = np.around(output[key], 1)
            else:
                output[key] = np.around(output[key], 2)

    return PHS(**output)


# Constants
const_t_eq = math.exp(-1 / 10)
const_t_sk = math.exp(-1 / 3)
const_sw = math.exp(-1 / 10)


@np.vectorize
@jit(nopython=True, cache=True)
def _phs_optimized(
    tdb,
    tr,
    v,
    p_a,
    met,
    clo,
    posture,
    drink,
    acclimatized,
    weight,
    wme,
    i_mst,
    a_p,
    height,
    walk_sp,
    theta,
    duration,
    f_r,
    t_sk,
    t_cr,
    t_re,
    t_cr_eq,
    t_sk_t_cr_wg,
    sw_tot,
    model,
):
    # DuBois body surface area [m2]
    a_dubois = 0.202 * (weight**0.425) * (height**0.725)
    # specific heat of the body [J/kg/C/min]
    sp_heat = met_to_w_m2 * weight / a_dubois

    d_lim_t_re = 0  # maximum allowable exposure time for heat storage [min]
    # maximum allowable exposure time for water loss, mean subject [min]
    d_lim_loss_50 = 0
    # maximum allowable exposure time for water loss, 95 % of the working population [min]
    d_lim_loss_95 = 0

    if model == Models.iso_7933_2023.value:
        # 2023 standard only has one d_max value
        d_max_50 = (0.03 if drink == 0 else 0.05) * weight * 1000
        d_max_95 = (0.03 if drink == 0 else 0.05) * weight * 1000
    else:  # model == Models.iso_7933_2004.value:
        # maximum water loss to protect a mean subject [g]
        d_max_50 = 0.075 * weight * 1000
        # maximum water loss to protect 95 % of the working population [g]
        d_max_95 = 0.05 * weight * 1000

    sweat_rate = sw_tot

    # def_dir = 1 for unidirectional walking, def_dir = 0 for omni-directional walking
    def_dir = 1 if theta != 0 else 0
    def_speed = 0 if walk_sp == 0 else 1

    # radiating area dubois
    if posture == Postures.standing.value:
        a_r_du = 0.77
    elif posture == Postures.sitting.value:
        a_r_du = 0.7
    else:  # posture == Postures.crouching.value:
        a_r_du = 0.67

    # evaluation of the max sweat rate as a function of the metabolic rate
    if model == Models.iso_7933_2004.value:
        sw_max = (met - 32) * a_dubois
        sw_max = min(sw_max, 400)
        sw_max = max(sw_max, 250)
        if acclimatized == 100:
            sw_max = sw_max * 1.25
    else:  # model == Models.iso_7933_2023.value:
        sw_max = 400 if acclimatized == 0 else 500

    # max skin wettedness
    w_max = 0.85 if acclimatized == 0 else 1

    # static clothing insulation
    i_cl_st = clo * 0.155

    fcl = 1 + 0.28 * clo if model == Models.iso_7933_2023.value else 1 + 0.3 * clo

    # Static boundary layer thermal insulation in quiet air
    i_a_st = 0.111

    # Total static insulation
    i_tot_st = i_cl_st + i_a_st / fcl
    if def_speed > 0:
        if def_dir == 1:  # Unidirectional walking
            v_r = abs(v - walk_sp * math.cos(3.14159 * theta / 180))
        elif v < walk_sp:
            v_r = walk_sp
        else:
            v_r = v
    else:
        walk_sp = 0.0052 * (met - 58)
        walk_sp = min(walk_sp, 0.7)
        v_r = v

    # Dynamic clothing insulation - correction for wind (Var) and walking speed
    v_ux = v_r
    if v_r > 3:
        v_ux = 3
    w_a_ux = walk_sp
    if walk_sp > 1.5:
        w_a_ux = 1.5
    # correction for the dynamic total dry thermal insulation at or above 0.6 clo
    corr_cl = 1.044 * math.exp(
        (0.066 * v_ux - 0.398) * v_ux + (0.094 * w_a_ux - 0.378) * w_a_ux,
    )
    corr_cl = min(corr_cl, 1)
    # correction for the dynamic total dry thermal insulation at 0 clo
    corr_ia = math.exp((0.047 * v_r - 0.472) * v_r + (0.117 * w_a_ux - 0.342) * w_a_ux)
    corr_ia = min(corr_ia, 1)
    corr_tot = corr_cl
    if clo <= 0.6:
        corr_tot = ((0.6 - clo) * corr_ia + clo * corr_cl) / 0.6
    # total dynamic clothing insulation
    i_tot_dyn = i_tot_st * corr_tot
    # dynamic boundary layer thermal insulation
    i_a_dyn = corr_ia * i_a_st
    i_cl_dyn = i_tot_dyn - i_a_dyn / fcl
    # correction for the dynamic permeability index
    corr_e = (2.6 * corr_tot - 6.5) * corr_tot + 4.9
    im_dyn = i_mst * corr_e
    im_dyn = min(im_dyn, 0.9)
    r_t_dyn = i_tot_dyn / im_dyn / 16.7
    t_exp = 28.56 + 0.115 * tdb + 0.641 * p_a  # expired air temperature
    # respiratory convective heat flow [W/m2]
    c_res = 0.001516 * met * (t_exp - tdb)
    # respiratory evaporative heat flow [W/m2]
    e_res = 0.00127 * met * (59.34 + 0.53 * tdb - 11.63 * p_a)
    z = 3.5 + 5.2 * v_r
    if v_r > 1:
        z = 8.7 * v_r**0.6

    # dynamic convective heat transfer coefficient
    if model == Models.iso_7933_2004.value:
        hc_dyn = 2.38 * abs(t_sk - tdb) ** 0.25
    else:  # model == Models.iso_7933_2023.value:
        t_cl = tr + 0.1  # clothing surface temperature
        hc_dyn = 2.38 * abs(t_cl - tdb) ** 0.25

    hc_dyn = max(hc_dyn, z)

    aux_r = 5.67e-08 * a_r_du

    if model == Models.iso_7933_2023.value:
        f_cl_r = (1 - a_p) * 0.97 + a_p * (1 - f_r)
    else:  # model == Models.iso_7933_2004.value:
        f_cl_r = (1 - a_p) * 0.97 + a_p * f_r

    for time in range(1, duration + 1):
        t_sk0 = t_sk
        t_re0 = t_re
        t_cr0 = t_cr
        t_cr_eq0 = t_cr_eq
        t_sk_t_cr_wg0 = t_sk_t_cr_wg

        # equilibrium core temperature associated to the metabolic rate
        t_cr_eq_m = 0.0036 * met + 36.6
        # Core temperature at this minute, by exponential averaging
        t_cr_eq = t_cr_eq0 * const_t_eq + t_cr_eq_m * (1 - const_t_eq)
        # Heat storage associated with this core temperature increase during the last minute
        d_stored_eq = sp_heat * (t_cr_eq - t_cr_eq0) * (1 - t_sk_t_cr_wg0)
        # skin temperature prediction -- clothed model
        t_sk_eq_cl = 12.165 + 0.02017 * tdb + 0.04361 * tr + 0.19354 * p_a - 0.25315 * v
        t_sk_eq_cl = t_sk_eq_cl + 0.005346 * met + 0.51274 * t_re
        # nude model
        t_sk_eq_nu = 7.191 + 0.064 * tdb + 0.061 * tr + 0.198 * p_a - 0.348 * v
        t_sk_eq_nu = t_sk_eq_nu + 0.616 * t_re
        if clo >= 0.6:
            t_sk_eq = t_sk_eq_cl
        elif clo <= 0.2:
            t_sk_eq = t_sk_eq_nu
        else:
            t_sk_eq = t_sk_eq_nu + 2.5 * (t_sk_eq_cl - t_sk_eq_nu) * (clo - 0.2)

        # skin temperature [C]
        t_sk = t_sk0 * const_t_sk + t_sk_eq * (1 - const_t_sk)
        # Saturated water vapour pressure at the surface of the skin
        p_sk = 0.6105 * math.exp(17.27 * t_sk / (t_sk + 237.3))
        t_cl = tr + 0.1  # clothing surface temperature
        while True:
            # radiative heat transfer coefficient
            h_r = f_cl_r * aux_r * ((t_cl + 273) ** 4 - (tr + 273) ** 4) / (t_cl - tr)
            t_cl_new = (fcl * (hc_dyn * tdb + h_r * tr) + t_sk / i_cl_dyn) / (
                fcl * (hc_dyn + h_r) + 1 / i_cl_dyn
            )
            if abs(t_cl - t_cl_new) <= 0.001:
                break
            t_cl = (t_cl + t_cl_new) / 2

        convection = fcl * hc_dyn * (t_cl - tdb)
        radiation = fcl * h_r * (t_cl - tr)
        # maximum evaporative heat flow at the skin surface [W/m2]
        e_max = (p_sk - p_a) / r_t_dyn
        if e_max == 0:  # added this otherwise e_req / e_max cannot be calculated
            e_max = 0.001
        # required evaporative heat flow [W/m2]
        e_req = met - d_stored_eq - wme - c_res - e_res - convection - radiation
        # required skin wettedness
        w_req = e_req / e_max

        if e_req <= 0:
            e_req = 0
            sw_req = 0  # required sweat rate [W/m2]
        elif e_max <= 0:
            e_max = 0
            sw_req = sw_max
        elif w_req >= 1.7:
            sw_req = sw_max
        else:
            e_v_eff = (2 - w_req) ** 2 / 2 if w_req > 1 else 1 - w_req**2 / 2

            e_v_eff = max(0.05, e_v_eff)

            sw_req = e_req / e_v_eff
            sw_req = min(sw_req, sw_max)
        sweat_rate = sweat_rate * const_sw + sw_req * (1 - const_sw)

        if sweat_rate <= 0:
            e_p = 0  # predicted evaporative heat flow [W/m2]
            sweat_rate = 0
        else:
            k = e_max / sweat_rate
            wp = 1
            if k >= 0.5:
                wp = -k + math.sqrt(k * k + 2)
            wp = min(wp, w_max)
            e_p = wp * e_max

        # body heat storage rate [W/m2]
        d_storage = e_req - e_p + d_stored_eq
        t_cr_new = t_cr0
        while True:
            t_sk_t_cr_wg = 0.3 - 0.09 * (t_cr_new - 36.8)
            t_sk_t_cr_wg = min(t_sk_t_cr_wg, 0.3)
            t_sk_t_cr_wg = max(t_sk_t_cr_wg, 0.1)
            t_cr = (
                d_storage / sp_heat
                + t_sk0 * t_sk_t_cr_wg0 / 2
                - t_sk * t_sk_t_cr_wg / 2
            )
            t_cr = (t_cr + t_cr0 * (1 - t_sk_t_cr_wg0 / 2)) / (1 - t_sk_t_cr_wg / 2)
            if abs(t_cr - t_cr_new) <= 0.001:
                break
            t_cr_new = (t_cr_new + t_cr) / 2

        t_re = t_re0 + (2 * t_cr - 1.962 * t_re0 - 1.31) / 9
        if d_lim_t_re == 0 and t_re >= 38:
            d_lim_t_re = time
        sw_tot = sw_tot + sweat_rate + e_res
        sw_tot_g = sw_tot * 2.67 * a_dubois / 1.8 / 60
        if d_lim_loss_50 == 0 and sw_tot_g >= d_max_50:
            d_lim_loss_50 = time
        if d_lim_loss_95 == 0 and sw_tot_g >= d_max_95:
            d_lim_loss_95 = time

    # in the standard the if statement is within the while loop, causing it to decay exponentially
    if drink == 0:
        d_lim_loss_95 = d_lim_loss_95 * 0.6
        d_lim_loss_50 = d_lim_loss_95
    if d_lim_loss_50 == 0:
        d_lim_loss_50 = duration
    if d_lim_loss_95 == 0:
        d_lim_loss_95 = duration
    if d_lim_t_re == 0:
        d_lim_t_re = duration

    return (
        t_re,
        t_sk,
        t_cr,
        t_cr_eq,
        t_sk_t_cr_wg,
        sweat_rate,
        sw_tot_g,
        d_lim_loss_50,
        d_lim_loss_95,
        d_lim_t_re,
    )



import numpy as np
import pytest

from pythermalcomfort.utilities import (
    PsychrometricValues,
    dew_point_tmp,
    enthalpy_air,
    mean_radiant_tmp,
    operative_tmp,
    p_sat,
    psy_ta_rh,
    wet_bulb_tmp,
)
from tests.conftest import is_equal


def test_t_dp() -> None:
    """Test the dew point temperature function with various inputs."""
    assert dew_point_tmp(31.6, 59.6) == 22.6
    assert dew_point_tmp(29.3, 75.4) == 24.3
    assert dew_point_tmp(27.1, 66.4) == 20.2


def test_t_wb() -> None:
    """Test the wet bulb temperature function with various inputs."""
    assert wet_bulb_tmp(27.1, 66.4) == 22.4
    assert wet_bulb_tmp(25, 50) == 18.0


def test_enthalpy() -> None:
    """Test the enthalpy function with various inputs."""
    assert is_equal(enthalpy_air(25, 0.01), 50561.25, 0.1)
    assert is_equal(enthalpy_air(27.1, 0.01), 52707.56, 0.1)


def test_psy_ta_rh() -> None:
    """Test the psychrometric function with various inputs."""
    assert psy_ta_rh(25, 50, p_atm=101325) == PsychrometricValues(
        p_sat=pytest.approx(3169.2, abs=1e-1),
        p_vap=pytest.approx(1584.6, abs=1e-1),
        hr=pytest.approx(0.009881547577511219, abs=1e-7),
        wet_bulb_tmp=18.0,
        dew_point_tmp=13.8,
        h=pytest.approx(50259.79, abs=1e-1),
    )


def test_t_o() -> None:
    """Test the operative temperature function with various inputs."""
    assert operative_tmp(25, 25, 0.1) == 25
    assert round(operative_tmp(25, 30, 0.3), 2) == 26.83
    assert round(operative_tmp(20, 30, 0.3), 2) == 23.66
    assert operative_tmp(25, 25, 0.1, standard="ASHRAE") == 25
    assert operative_tmp(20, 30, 0.1, standard="ASHRAE") == 25
    assert operative_tmp(20, 30, 0.3, standard="ASHRAE") == 24
    assert operative_tmp(20, 30, 0.7, standard="ASHRAE") == 23


def test_p_sat() -> None:
    """Test the saturation pressure function with various inputs."""
    assert pytest.approx(p_sat(tdb=25), abs=1e-1) == 3169.2
    assert pytest.approx(p_sat(tdb=50), abs=1e-1) == 12349.9


def test_t_mrt() -> None:
    """Test the mean radiant temperature function with various inputs."""
    np.testing.assert_equal(
        mean_radiant_tmp(
            tg=[53.2, 55, 55],
            tdb=30,
            v=[0.3, 0.3, 0.1],
            d=0.1,
            standard="ISO",
        ),
        [74.8, 77.8, 71.9],
    )
    np.testing.assert_equal(
        mean_radiant_tmp(
            tg=[25.42, 26.42, 26.42, 26.42],
            tdb=26.10,
            v=0.1931,
            d=[0.1, 0.1, 0.5, 0.03],
            standard="Mixed Convection",
        ),
        [24.2, 27.0, np.nan, np.nan],
    )



from __future__ import annotations

import numpy as np
import pytest

from pythermalcomfort.classes_return import WorkCapacity
from pythermalcomfort.models.work_capacity_iso import work_capacity_iso


def _expected_capacity(
    wbgt: float | list[float],
    met: float | list[float],
) -> np.ndarray:
    """Calculate the expected work capacity based on WBGT and metabolic rate."""
    wbgt_arr = np.array(wbgt)
    met_arr = np.array(met)
    met_rest = 117.0
    wbgt_lim = 34.9 - met_arr / 46.0
    wbgt_lim_rest = 34.9 - met_rest / 46.0
    cap = ((wbgt_lim_rest - wbgt_arr) / (wbgt_lim_rest - wbgt_lim)) * 100.0
    return np.clip(cap, 0, 100)


def test_scalar_typical() -> None:
    """Test that the function returns a WorkCapacity object with correct capacity."""
    wbgt = 30.0
    met = 200.0
    result = work_capacity_iso(wbgt, met)
    assert isinstance(result, WorkCapacity)
    assert isinstance(result.capacity, float)
    expected = _expected_capacity(wbgt, met)
    assert pytest.approx(expected, rel=1e-3) == result.capacity


def test_list_input_pairwise() -> None:
    """Test that the function handles list inputs correctly."""
    wbgts = [20.0, 30.0, 40.0]
    mets = [100.0, 200.0, 300.0]
    result = work_capacity_iso(wbgts, mets)
    assert isinstance(result.capacity, np.ndarray)
    expected = _expected_capacity(wbgts, mets)
    for got, exp in zip(result.capacity, expected, strict=False):
        assert pytest.approx(exp, rel=1e-3) == got


def test_exact_wbgt_lim_full_capacity() -> None:
    """Test that the function returns full capacity when WBGT limit is met."""
    met = 250.0
    wbgt_lim = 34.9 - met / 46.0
    result = work_capacity_iso(wbgt_lim, met)
    assert result.capacity == pytest.approx(100.0, abs=1e-6)


def test_resting_limit_zero_capacity() -> None:
    """Test that the function returns zero capacity when WBGT limit for resting is met."""
    met = 250.0
    wbgt_lim_rest = 34.9 - 117.0 / 46.0
    result = work_capacity_iso(wbgt_lim_rest, met)
    assert result.capacity == pytest.approx(0.0, abs=1e-6)


def test_low_wbgt_clamped_to_100() -> None:
    """Test that the function clamps low WBGT values to 100% capacity."""
    result = work_capacity_iso(0.0, 500.0)
    assert result.capacity == pytest.approx(100.0, abs=1e-6)


def test_high_wbgt_clamped_to_0() -> None:
    """Test that the function clamps high WBGT values to 0% capacity."""
    result = work_capacity_iso(100.0, 500.0)
    assert result.capacity == pytest.approx(0.0, abs=1e-6)


def test_negative_met_raises() -> None:
    """Test that the function raises ValueError for negative met."""
    with pytest.raises(ValueError):
        work_capacity_iso(25.0, -10.0)


def test_met_above_max_raises() -> None:
    """Test that the function raises ValueError for met above maximum."""
    with pytest.raises(ValueError):
        work_capacity_iso(25.0, 3000.0)



import os

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd

from pythermalcomfort.jos3_functions.parameters import local_clo_typical_ensembles
from pythermalcomfort.models import JOS3

# Make "jos3_example_example" directory in the current directory
directory_name = "jos3_output_example"
current_directory = os.getcwd()
jos3_example_directory = os.path.join(current_directory, directory_name)
if not os.path.exists(jos3_example_directory):
    os.makedirs(jos3_example_directory)

# -------------------------------------------
# EXAMPLE 1 (simple simulation)
# -------------------------------------------


def simple_simulation():
    """Run a simple simulation example."""
    # Build a model and set a body built
    # Create an instance of the JOS3 class with optional body parameters such as body height, weight, age, sex, etc.
    model = JOS3(height=1.7, weight=60, age=30)

    # Set the first phase
    model.to = 28  # Operative temperature [°C]
    model.rh = 40  # Relative humidity [%]
    model.v = 0.2  # Air velocity [m/s]
    model.par = 1.2  # Physical activity ratio [-]
    model.simulate(60)  # Exposure time = 60 [min]

    # Set the next condition (You only need to change the parameters that you want to change)
    model.to = 20  # Change only operative temperature
    model.simulate(60)  # Additional exposure time = 60 [min]

    # Show the results
    df = pd.DataFrame(model.dict_results())  # Make pandas.DataFrame
    df.t_skin_mean.plot()  # Plot time series of mean skin temperature.
    plt.ylabel(
        "Mean skin temperature [°C]",
    )  # Set y-label as 'Mean skin temperature [°C]'
    plt.xlabel("Time [min]")  # Set x-label as 'Time [min]'
    plt.savefig(
        os.path.join(jos3_example_directory, "jos3_example1_mean_skin_temperature.png"),
    )  # Save plot at the current directory
    plt.show()  # Show the plot

    # Exporting the results as csv
    model.to_csv(
        os.path.join(jos3_example_directory, "jos3_example1 (default output).csv"),
    )

    # Print the BMR value using the getter
    print("BMR=", model.bmr)
    print("Body name list: ", model.body_names)


def detailed_simulation():
    """Run a detailed simulation example with multiple phases."""
    # Build a model and set a body built
    # Create an instance of the JOS3 class with optional body parameters such as body height, weight, age, sex, etc.
    model = JOS3(
        height=1.7,
        weight=60,
        fat=20,
        age=30,
        sex="male",
        bmr_equation="japanese",
        bsa_equation="fujimoto",
    )

    # Set environmental conditions such as air temperature, mean radiant temperature using the setter methods.
    # Set the first condition
    # Environmental parameters can be input as int, float, list, dict, numpy array format.
    model.tdb = 28  # Air temperature [°C]
    model.tr = 30  # Mean radiant temperature [°C]
    model.rh = 40  # Relative humidity [%]
    model.v = np.array(  # Air velocity [m/s]
        [
            0.2,  # head
            0.4,  # neck
            0.4,  # chest
            0.1,  # back
            0.1,  # pelvis
            0.4,  # left shoulder
            0.4,  # left arm
            0.4,  # left hand
            0.4,  # right shoulder
            0.4,  # right arm
            0.4,  # right hand
            0.1,  # left thigh
            0.1,  # left leg
            0.1,  # left foot
            0.1,  # right thigh
            0.1,  # right leg
            0.1,  # right foot
        ],
    )
    model.clo = local_clo_typical_ensembles[
        "briefs, socks, undershirt, work jacket, work pants, safety shoes"
    ]["local_body_part"]

    # par should be input as int, float.
    model.par = 1.2  # Physical activity ratio [-], assuming a sitting position
    # posture should be input as int (0, 1, or 2) or str ("standing", "sitting" or "lying").
    # (0="standing", 1="sitting" or 2="lying")
    model.posture = "sitting"  # Posture [-], assuming a sitting position

    # Run JOS-3 model
    model.simulate(
        times=30,  # Number of loops of a simulation
        dtime=60,  # Time delta [sec]. The default is 60.
    )  # Exposure time = 30 [loops] * 60 [sec] = 30 [min]

    # Set the next condition (You only need to change the parameters that you want to change)
    model.to = 20  # Change operative temperature
    model.v = {  # Air velocity [m/s], assuming to use a desk fan
        "head": 0.2,
        "neck": 0.4,
        "chest": 0.4,
        "back": 0.1,
        "pelvis": 0.1,
        "left_shoulder": 0.4,
        "left_arm": 0.4,
        "left_hand": 0.4,
        "right_shoulder": 0.4,
        "right_arm": 0.4,
        "right_hand": 0.4,
        "left_thigh": 0.1,
        "left_leg": 0.1,
        "left_foot": 0.1,
        "right_thigh": 0.1,
        "right_leg": 0.1,
        "right_foot": 0.1,
    }
    # Run JOS-3 model
    model.simulate(
        times=60,  # Number of loops of a simulation
        dtime=60,  # Time delta [sec]. The default is 60.
    )  # Additional exposure time = 60 [loops] * 60 [sec] = 60 [min]

    # Set the next condition (You only need to change the parameters that you want to change)
    model.tdb = 30  # Change air temperature [°C]
    model.tr = 35  # Change mean radiant temperature [°C]
    # Run JOS-3 model
    model.simulate(
        times=30,  # Number of loops of a simulation
        dtime=60,  # Time delta [sec]. The default is 60.
    )  # Additional exposure time = 30 [loops] * 60 [sec] = 30 [min]

    # Show the results
    df = pd.DataFrame(model.dict_results())  # Make pandas.DataFrame
    df[
        ["t_skin_mean", "t_skin_head", "t_skin_chest", "t_skin_left_hand"]
    ].plot()  # Plot time series of local skin temperature.
    plt.legend(["Mean", "Head", "Chest", "Left hand"])  # Reset the legends
    plt.ylabel("Skin temperature [°C]")  # Set y-label as 'Skin temperature [°C]'
    plt.xlabel("Time [min]")  # Set x-label as 'Time [min]'
    plt.savefig(
        os.path.join(jos3_example_directory, "jos3_example2_skin_temperatures.png"),
    )  # Save plot at the current directory
    plt.show()  # Show the plot

    # Exporting the results as csv
    model.to_csv(os.path.join(jos3_example_directory, "jos3_example2 (all output).csv"))


def female_simulation():
    """Run a simple simulation example."""
    model = JOS3(height=1.7, weight=60, age=30, sex="female", bmr_equation="japanese")

    # Set the first phase
    model.to = 28
    model.rh = 40
    model.v = 0.2
    model.par = 1.2
    model.clo = 0.3
    model.simulate(60)

    # Set the next condition
    model.to = 20
    model.simulate(60)

    # Show the results
    df = pd.DataFrame(model.dict_results())
    df.t_skin_mean.plot()
    plt.ylabel("Mean skin temperature [°C]")
    plt.xlabel("Time [min]")
    plt.savefig(
        os.path.join(jos3_example_directory, "jos3_example3_mean_skin_temperature.png"),
    )
    plt.show()

    # Exporting the results as csv
    model.to_csv(
        os.path.join(jos3_example_directory, "jos3_example3 (female simulation).csv"),
    )


def validation_simulation():
    """Run validation simulation using Stolwijk and Hardy dataset."""
    """
    Following code is for validation between experimental and predicted data
    """
    exp_dataset_name = "human_subject_experiment_dataset.xlsx"
    exp_dataset_path = os.path.join(jos3_example_directory, exp_dataset_name)

    # Initialize an empty dictionary to hold the datasets
    exp_dataset = {}

    # List of sheet names and their respective header row indices to be read
    sheet_names = [("Stolwijk1966", 0), ("Werner1980", 0)]

    try:
        # Loop through each sheet name and read the data into a DataFrame
        for sheet_name, header in sheet_names:
            exp_dataset[sheet_name] = pd.read_excel(
                exp_dataset_path,
                header=header,
                sheet_name=sheet_name,
            )
    # Handle the case where the file is not found
    except FileNotFoundError:
        print(f"File {exp_dataset_path} not found.")
    # Handle other general exceptions
    except Exception as e:
        print(f"An error occurred: {e}")

    # Concatenate all the individual data frames into a single DataFrame
    sim_dataset = {}

    def sim_stolwijk_hardy(models, tolist, rhlist):
        result = []
        for model in models:
            model.icl = np.array(
                [
                    0,
                    0,
                    0,
                    0.3,
                    0.3,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0.3,
                    0.3,
                    0,
                    0.3,
                    0.3,
                    0,
                ],
            )
            model.par = 1.2
            model.posture = "sitting"
            # model.options["nonshivering_thermogenesis"] = True
            # model.options["limit_dshiv/dt"] = True
            # model.options["shivering_threshold"] = True

            model.to = tolist[0]
            model.rh = rhlist[0]
            model.simulate(10, dtime=600)

            model.to = tolist[1]
            model.rh = rhlist[1]
            model.simulate(60)

            model.to = tolist[2]
            model.rh = rhlist[2]
            model.simulate(120)

            model.to = tolist[3]
            model.rh = rhlist[3]
            model.simulate(60)

            sim = pd.DataFrame(model.dict_results())
            sim = sim.loc[
                10:,
                "t_skin_mean":,
            ]
            result.append(sim.copy())

        avgsim = (result[0] + result[1] + result[2]) / 3
        avgsim = avgsim.iloc[::5]
        avgsim.index = [i for i in range(0, 241, 5)]
        avgsim["time"] = [i for i in range(0, 241, 5)]

        return avgsim.copy()

    simdatas = {}

    # -------------------------------------------------------------------------------
    # A : Stolwijk and Hardy, 1966
    # J.A.J. Stolwijk, J.D. Hardy, Partitional calorimetric studies of responses of man
    # to thermal transients, J. Appl. Physiol. 21 (3) (1966) 967–977,
    # https://doi.org/10.1152/jappl.1966.21.6.1799.
    # -------------------------------------------------------------------------------
    dnames = ["A-FIG.4", "A-FIG.5", "A-FIG.6", "A-FIG.7"]

    inputcons = {
        "A-FIG.4": {"to": [28.0, 27.8, 33.3, 28.0], "rh": [40, 37, 34, 37]},
        "A-FIG.5": {"to": [28.0, 28.5, 37.5, 28.5], "rh": [40, 40, 34, 41]},
        "A-FIG.6": {"to": [28.0, 28.0, 42.5, 28.1], "rh": [40, 37, 34, 37]},
        "A-FIG.7": {"to": [28.0, 28.1, 47.8, 28.3], "rh": [40, 43, 27, 44]},
    }

    for d in dnames:
        models = []
        models.append(
            JOS3(
                height=1.95,
                weight=88.6,
                age=25,
            ),
        )
        models.append(
            JOS3(
                height=1.84,
                weight=76.1,
                age=22,
            ),
        )
        models.append(
            JOS3(
                height=1.75,
                weight=110,
                age=23,
            ),
        )

        tolist = inputcons[d]["to"]
        rhlist = inputcons[d]["rh"]
        output = sim_stolwijk_hardy(models, tolist, rhlist)
        print("simulated:", d)

        simdatas[d] = output

    # -------------------------------------------------------------------------------
    # B : Stolwijk and Hardy, 1966
    #  J.D. Hardy, J.A.J. Stolwjk, Partitional exposures calorimetric studies of man
    # during exposures to thermal transients, J. Appl. Physiol. 21 (1966) 1799–1806,
    # https://doi.org/10.1152/jappl.1966.21.6.1799.
    # -------------------------------------------------------------------------------
    dnames = ["B-FIG.1", "B-FIG.2", "B-FIG.3", "B-FIG.4", "B-FIG.5"]

    inputcons = {
        "B-FIG.1": {"to": [28.0, 29.0, 22.0, 29.0], "rh": [40, 44, 39, 41]},
        "B-FIG.2": {"to": [28.0, 28.0, 18.0, 28.0], "rh": [40, 40, 40, 40]},
        "B-FIG.3": {"to": [28.0, 22.3, 43.5, 22.6], "rh": [40, 40, 38, 36]},
        "B-FIG.4": {"to": [28.0, 18.0, 42.0, 18.0], "rh": [40, 40, 40, 40]},
        "B-FIG.5": {"to": [28.0, 43.0, 17.0, 43.0], "rh": [40, 40, 40, 40]},
    }

    for d in dnames:
        models = []
        models.append(
            JOS3(
                height=1.91,
                weight=77.2,
                age=25,
            ),
        )
        models.append(
            JOS3(
                height=1.91,
                weight=84.5,
                age=26,
            ),
        )
        models.append(
            JOS3(
                height=1.88,
                weight=92.7,
                age=22,
            ),
        )

        tolist = inputcons[d]["to"]
        rhlist = inputcons[d]["rh"]
        output = sim_stolwijk_hardy(models, tolist, rhlist)
        print("simulated:", d)

        simdatas[d] = output

    to_concat = []
    for key, value in simdatas.items():
        value["Condition"] = key
        to_concat.append(value.copy())
    df = pd.concat(to_concat).reset_index(drop=True)

    sim_dataset["Stolwijk1966"] = df.copy()

    gexp = exp_dataset["Stolwijk1966"].copy()
    gsim = sim_dataset["Stolwijk1966"].copy()
    dfdict = {}
    for dn in [
        "A-FIG.4",
        "A-FIG.5",
        "A-FIG.6",
        "A-FIG.7",
        "B-FIG.1",
        "B-FIG.2",
        "B-FIG.3",
        "B-FIG.4",
        "B-FIG.5",
    ]:
        dfdict[dn] = pd.DataFrame(
            {
                "TreExp": gexp.loc[gexp["Condition"] == dn, "Tre"],
                "TreSim": gsim.loc[gsim["Condition"] == dn, "t_core_pelvis"],
                "TskExp": gexp.loc[gexp["Condition"] == dn, "Tsk"],
                "TskSim": gsim.loc[gsim["Condition"] == dn, "t_skin_mean"],
            },
        )

    dfs = []
    for key, value in dfdict.items():
        for seg in ["Tre", "Tsk"]:
            df = value[[seg + "Exp", seg + "Sim"]].copy()
            df.columns = ["Exp", "Sim"]
            df["Time"] = [i for i in range(0, 241, 5)]
            df["ExpName"] = key
            df["Type"] = seg
            dfs.append(df.copy())

    def plotdata(ser, ax, color, markersize=7, legend=False):
        mk = None
        me = 1
        if color == "black":
            ser.plot(
                ax=ax,
                linewidth=2,
                color=(0.0, 0.0, 0.0, 0.7),
                marker="o",
                markersize=markersize,
                legend=legend,
                markerfacecolor=mk,
                markeredgewidth=me,
            )
        elif color == "blue":
            ser.plot(
                ax=ax,
                linewidth=2,
                color=(0.1, 0.3, 1.0, 0.7),
                marker="^",
                markersize=markersize,
                legend=legend,
                markerfacecolor=mk,
                markeredgewidth=me,
            )
        elif color == "green":
            ser.plot(
                ax=ax,
                linewidth=2,
                color=(0.2, 0.6, 0.2, 0.7),
                marker="s",
                markersize=markersize,
                legend=legend,
                markerfacecolor=mk,
                markeredgewidth=me,
            )
        elif color == "gray":
            ser.plot(
                ax=ax,
                linewidth=2,
                color=(0.3, 0.3, 0.3, 0.7),
                marker="o",
                markersize=markersize,
                legend=legend,
                markerfacecolor=mk,
                markeredgewidth=me,
            )
        elif color == "lightblue":
            ser.plot(
                ax=ax,
                linewidth=2,
                color=(0.3, 0.1, 0.7, 0.7),
                marker="^",
                markersize=markersize,
                legend=legend,
                markerfacecolor=mk,
                markeredgewidth=me,
            )

    def graph(gexp, gsim, ax):
        gexp.index = [i for i in range(0, 241, 5)]
        gsim.index = [i for i in range(0, 241, 5)]

        # Tre
        ser = gexp["Tre"]
        ser.name = "Rectum (EXP)"
        plotdata(ser, ax, "black")
        ser = gsim["t_core_pelvis"]
        ser.name = "Rectum (JOS-3)"
        plotdata(ser, ax, "blue")

        # Tsk
        ser = gexp["Tsk"]
        ser.name = "Mean skin (EXP)"
        plotdata(ser, ax, "gray")
        ser = gsim["t_skin_mean"]
        ser.name = "Mean skin (JOS-3)"
        plotdata(ser, ax, "lightblue")

        ax.set_ylim((28, 40))
        ax.set_yticks([i for i in range(28, 41, 2)])
        ax.set_xlim((0, 240))
        ax.set_xticks([i for i in range(0, 241, 30)])

    inputcons = {
        "A-FIG.4": {"to": [28.0, 27.8, 33.3, 28.0], "rh": [40, 37, 34, 37]},
        "A-FIG.5": {"to": [28.0, 28.5, 37.5, 28.5], "rh": [40, 40, 34, 41]},
        "A-FIG.6": {"to": [28.0, 28.0, 42.5, 28.1], "rh": [40, 37, 34, 37]},
        "A-FIG.7": {"to": [28.0, 28.1, 47.8, 28.3], "rh": [40, 43, 27, 44]},
        "B-FIG.1": {"to": [28.0, 29.0, 22.0, 29.0], "rh": [40, 44, 39, 41]},
        "B-FIG.2": {"to": [28.0, 28.0, 18.0, 28.0], "rh": [40, 40, 40, 40]},
        "B-FIG.3": {"to": [28.0, 22.3, 43.5, 22.6], "rh": [40, 40, 38, 36]},
        "B-FIG.4": {"to": [28.0, 18.0, 42.0, 18.0], "rh": [40, 40, 40, 40]},
        "B-FIG.5": {"to": [28.0, 43.0, 17.0, 43.0], "rh": [40, 40, 40, 40]},
    }

    gexp = exp_dataset["Stolwijk1966"].copy()
    gsim = sim_dataset["Stolwijk1966"].copy()
    fig, axes = plt.subplots(
        nrows=5,
        ncols=2,
        sharex=True,
        sharey=True,
        figsize=(8, 12),
    )

    for i, dn in enumerate(
        [
            "A-FIG.4",
            "A-FIG.5",
            "A-FIG.6",
            "A-FIG.7",
            "B-FIG.1",
            "B-FIG.2",
            "B-FIG.3",
            "B-FIG.4",
            "B-FIG.5",
        ],
    ):
        # if i >= 4:
        i += 1
        yi = i % 5
        xi = i // 5
        ax = axes[yi, xi]
        graph(gexp.loc[gexp["Condition"] == dn], gsim.loc[gsim["Condition"] == dn], ax)

        ax.set_title(dn)

        ax.set_xticks([i for i in range(0, 241, 30)])
        ax.set_xticklabels([str(i) for i in range(0, 241, 30)])

        ax.axvspan(60, 180, color=(0, 0, 0, 0.1))
        ax.axhspan(39, 40, color=(1, 1, 1, 0.6))
        t = "{:.1f}°C".format(inputcons[dn]["to"][1])
        ax.text(30, 39.1, t, horizontalalignment="center", c="black", fontsize=10)
        t = "{:.1f}°C".format(inputcons[dn]["to"][2])
        ax.text(120, 39.1, t, horizontalalignment="center", c="black", fontsize=10)
        t = "{:.1f}°C".format(inputcons[dn]["to"][3])
        ax.text(210, 39.1, t, horizontalalignment="center", c="black", fontsize=10)

    fig.delaxes(axes[0, 0])
    handles, labels = ax.get_legend_handles_labels()
    fig.legend(
        handles,
        labels,
        loc="upper left",
        bbox_to_anchor=(0.1, 0.95),
        frameon=False,
        fontsize=12,
    )

    fig.subplots_adjust(left=0.1, right=0.9, top=0.9, bottom=0.1, hspace=0.4)

    fig.text(
        0.03,
        0.5,
        "Body temperature [°C]",
        va="center",
        rotation="vertical",
        fontsize=12,
    )
    fig.text(0.53, 0.03, "Time [min]", ha="center", fontsize=12)

    plt.tight_layout(rect=[0.05, 0.05, 1, 1])
    fig.savefig(
        os.path.join(jos3_example_directory, "jos3_validation_with_Stolwijk1966.png"),
    )


def main():
    """Run the selected simulation."""
    # Simulation switch (Please change as needed)
    run_simple_simulation = True
    run_detailed_simulation = True
    run_female_simulation = True
    run_validation_simulation = True

    if run_simple_simulation:
        simple_simulation()
    if run_detailed_simulation:
        detailed_simulation()
    if run_female_simulation:
        female_simulation()
    if run_validation_simulation:
        validation_simulation()


if __name__ == "__main__":
    main()



from __future__ import annotations

import warnings
from typing import Literal

import numpy as np
from scipy import optimize

from pythermalcomfort.classes_input import CEInputs
from pythermalcomfort.classes_return import CE
from pythermalcomfort.models.set_tmp import set_tmp
from pythermalcomfort.utilities import Units, units_converter


def cooling_effect(
    tdb: float | list[float],
    tr: float | list[float],
    vr: float | list[float],
    rh: float | list[float],
    met: float | list[float],
    clo: float | list[float],
    wme: float | list[float] = 0,
    units: Literal["SI", "IP"] = Units.SI.value,
) -> CE:
    """Return the value of the Cooling Effect (`CE`_) calculated in compliance
    with the ASHRAE 55 2020 Standard [55ASHRAE2023]_. The `CE`_ of the elevated air speed
    is the value that, when subtracted equally from both the average air
    temperature and the mean radiant temperature, yields the same `SET`_ under
    still air as in the first `SET`_ calculation under elevated air speed. The
    cooling effect is calculated only for air speed higher than 0.1 m/s.

    .. _CE: https://en.wikipedia.org/wiki/Thermal_comfort#Cooling_Effect
    .. _SET: https://en.wikipedia.org/wiki/Thermal_comfort#Standard_effective_temperature

    Parameters
    ----------
    tdb : float or list of floats
        Dry bulb air temperature, default in [°C] or [°F] if `units` = 'IP'.
    tr : float or list of floats
        Mean radiant temperature, default in [°C] or [°F] if `units` = 'IP'.
    vr : float or list of floats
        Relative air speed, default in [m/s] or [fps] if `units` = 'IP'.

        .. note::
            The cooling effect is calculated only for air speed higher than 0.1 m/s. If the air speed is lower than 0.1 m/s the function will return 0.

        .. note::
            vr is the relative air speed caused by body movement and not the air
            speed measured by the air speed sensor. The relative air speed is the sum of the
            average air speed measured by the sensor plus the activity-generated air speed
            (Vag). Where Vag is the activity-generated air speed caused by motion of
            individual body parts. vr can be calculated using the function
            :py:meth:`pythermalcomfort.utilities.v_relative`.

    rh : float or list of floats
        Relative humidity, [%].
    met : float or list of floats
        Metabolic rate, [met].
    clo : float or list of floats
        Clothing insulation, [clo].

        .. note::
            this is the basic insulation also known as the intrinsic clothing insulation value of the
            clothing ensemble (`I`:sub:`cl,r`), this is the thermal insulation from the skin
            surface to the outer clothing surface, including enclosed air layers, under actual
            environmental conditions. This value is not the total insulation (`I`:sub:`T,r`).
            The dynamic clothing insulation, clo, can be calculated using the function
            :py:meth:`pythermalcomfort.utilities.clo_dynamic_ashrae`.

    wme : float or list of floats, optional
        External work, [met]. Defaults to 0.
    units : {'SI', 'IP'}, optional
        Select the SI (International System of Units) or the IP (Imperial Units) system.
        Supported values are 'SI' and 'IP'. Defaults to 'SI'.

    Returns
    -------
    CE
        A dataclass containing the Cooling Effect value. See :py:class:`~pythermalcomfort.classes_return.CoolingEffect` for more details.
        To access the `ce` value, use the `ce` attribute of the returned `CoolingEffect` instance, e.g., `result.ce`.

    Notes
    -----
    .. note::
        Limitations:
        - This equation may not be accurate for extreme temperature ranges.

    Examples
    --------
    .. code-block:: python

        from pythermalcomfort.models import cooling_effect

        result = cooling_effect(tdb=25, tr=25, vr=0.3, rh=50, met=1.2, clo=0.5)
        print(result.ce)  # 1.68

        result = cooling_effect(
            tdb=[25, 77],
            tr=[25, 77],
            vr=[0.3, 1.64],
            rh=[50, 50],
            met=[1.2, 1],
            clo=[0.5, 0.6],
            units="IP",
        )
        print(result.ce)  # [0, 3.95]

    """
    # Validate inputs using the CoolingEffectInputs class
    CEInputs(
        tdb=tdb,
        tr=tr,
        vr=vr,
        rh=rh,
        met=met,
        clo=clo,
        wme=wme,
        units=units,
    )

    tdb = np.array(tdb)
    tr = np.array(tr)
    vr = np.array(vr)
    rh = np.array(rh)
    met = np.array(met)
    clo = np.array(clo)
    wme = np.array(wme)

    if units.upper() == Units.IP.value:
        tdb, tr, vr = units_converter(tdb=tdb, tr=tr, v=vr)

    still_air_threshold = 0.1

    _ce = _cooling_effect_vectorised(
        tdb=tdb,
        tr=tr,
        still_air_threshold=still_air_threshold,
        rh=rh,
        met=met,
        clo=clo,
        wme=wme,
        vr=vr,
    )

    if units.upper() == Units.IP.value:
        _ce = _ce / 1.8 * 3.28

    return CE(ce=np.around(_ce, 2))


@np.vectorize
def _cooling_effect_vectorised(tdb, tr, still_air_threshold, rh, met, clo, wme, vr):
    if vr <= 0.1:
        return 0.0

    initial_set_tmp = set_tmp(
        tdb=tdb,
        tr=tr,
        v=vr,
        rh=rh,
        met=met,
        clo=clo,
        wme=wme,
        round_output=False,
        calculate_ce=True,
        limit_inputs=False,
    ).set

    def function(x):
        return (
            set_tmp(
                tdb=tdb - x,
                tr=tr - x,
                v=still_air_threshold,
                rh=rh,
                met=met,
                clo=clo,
                wme=wme,
                round_output=False,
                calculate_ce=True,
                limit_inputs=False,
            ).set
            - initial_set_tmp
        )

    try:
        ce = optimize.brentq(function, 0.0, 40)
    except ValueError:
        ce = 0.0

    if ce == 0.0:
        warnings.warn(
            "Cooling effect could not be calculated. Returning 0.",
            UserWarning,
            stacklevel=2,
        )

    return ce



from __future__ import annotations

from dataclasses import dataclass, field
from enum import Enum

import numpy as np

from pythermalcomfort.utilities import Postures, Sex, Units, validate_type


class WorkIntensity(str, Enum):
    """Enumeration for work intensity levels."""

    HEAVY = "heavy"
    MODERATE = "moderate"
    LIGHT = "light"


@dataclass
class BaseInputs:
    """Base class containing all possible input parameters."""

    body_surface_area: float | int | np.ndarray | list = field(default=1.8258)
    tdb: float | int | np.ndarray | list = field(default=None)
    tr: float | int | np.ndarray | list = field(default=None)
    twb: float | int | np.ndarray | list = field(default=None)
    tg: float | int | np.ndarray | list = field(default=None)
    vr: float | int | np.ndarray | list = field(default=None)
    v: float | int | np.ndarray | list = field(default=None)
    rh: float | int | np.ndarray | list = field(default=None)
    met: float | int | np.ndarray | list = field(default=None)
    clo: float | int | np.ndarray | list = field(default=None)
    wme: float | int | np.ndarray | list = field(default=0)
    round_output: bool = field(default=True)
    limit_inputs: bool = field(default=True)
    with_solar_load: bool = field(default=False)
    airspeed_control: bool = field(default=True)
    units: str = field(default=Units.SI.value)
    a_coefficient: float | int = field(default=None)
    e_coefficient: float | int = field(default=None)
    v_ankle: float | int | np.ndarray | list = field(default=None)
    t_running_mean: float | int | np.ndarray | list = field(default=None)
    q: float | int | np.ndarray | list = field(default=None)
    tout: float | int | np.ndarray | list = field(default=None)
    p_atm: float | int | np.ndarray | list = field(default=101325)
    age: float | int | np.ndarray | list = field(default=None)
    weight: float | int | np.ndarray | list = field(default=None)
    height: float | int | np.ndarray | list = field(default=None)
    sol_altitude: float | int | np.ndarray | list = field(default=None)
    sharp: float | int | np.ndarray | list = field(default=None)
    sol_radiation_dir: float | int | np.ndarray | list = field(default=None)
    sol_radiation_global: float | int | np.ndarray | list = field(default=None)
    sol_transmittance: float | int | np.ndarray | list = field(default=None)
    f_svv: float | int | np.ndarray | list = field(default=None)
    f_bes: float | int | np.ndarray | list = field(default=None)
    asw: float | int | np.ndarray | list = field(default=None)
    floor_reflectance: float | int | np.ndarray | list = field(default=None)
    vertical_tmp_grad: float | int | np.ndarray | list = field(default=None)
    position: str | np.ndarray | list = field(default=None)
    sex: str | np.ndarray | list = field(default=None)
    posture: str | np.ndarray | list = field(default=None)
    max_skin_blood_flow: float | int | np.ndarray | list = field(default=80)
    max_sweating: float | int | np.ndarray | list = field(default=500)
    w_max: float | int | np.ndarray | list = field(default=None)
    wbgt: float | int | np.ndarray | list = field(default=None)
    work_intensity: str | WorkIntensity = field(default=None)
    thickness_quilt: float | int | np.ndarray | list = field(default=None)
    vapor_pressure: float | int | np.ndarray | list = field(default=None)

    def __post_init__(self):
        def is_pandas_series(obj):
            return type(obj).__name__ == "Series"

        def convert_series_to_list(obj):
            return obj.tolist() if is_pandas_series(obj) else obj

        def _validate_str_values(name: str, value, allowed):
            values = np.atleast_1d(value)
            for val in values.astype(str):
                if val.lower() not in allowed:
                    error_msg = f"{name} must be one of {allowed!r}, "
                    raise ValueError(error_msg)

        # Only validate attributes that are not None
        if self.units.upper() not in [Units.SI.value, Units.IP.value]:
            raise ValueError("Units must be either 'SI' or 'IP'")
        if self.position is not None:
            _validate_str_values(
                "position",
                self.position,
                [
                    Postures.sitting.value,
                    Postures.standing.value,
                    "standing, forced convection",
                ],
            )
        if self.posture is not None:
            _validate_str_values(
                "posture",
                self.posture,
                [
                    Postures.sitting.value,
                    Postures.standing.value,
                    Postures.crouching.value,
                ],
            )
        if self.sex is not None:
            _validate_str_values("sex", self.sex, [Sex.male.value, Sex.female.value])
        if self.tdb is not None:
            self.tdb = convert_series_to_list(self.tdb)
            validate_type(self.tdb, "tdb", (float, int, np.ndarray, list))
        if self.tr is not None:
            self.tr = convert_series_to_list(self.tr)
            validate_type(self.tr, "tr", (float, int, np.ndarray, list))
        if self.twb is not None:
            self.twb = convert_series_to_list(self.twb)
            validate_type(self.twb, "twb", (float, int, np.ndarray, list))
        if self.tg is not None:
            self.tg = convert_series_to_list(self.tg)
            validate_type(self.tg, "tg", (float, int, np.ndarray, list))
        if self.vr is not None:
            self.vr = convert_series_to_list(self.vr)
            validate_type(self.vr, "vr", (float, int, np.ndarray, list))
        if self.v is not None:
            self.v = convert_series_to_list(self.v)
            validate_type(self.v, "v", (float, int, np.ndarray, list))
        if self.rh is not None:
            self.rh = convert_series_to_list(self.rh)
            validate_type(self.rh, "rh", (float, int, np.ndarray, list))
        if self.met is not None:
            self.met = convert_series_to_list(self.met)
            validate_type(self.met, "met", (float, int, np.ndarray, list))
        if self.clo is not None:
            self.clo = convert_series_to_list(self.clo)
            validate_type(self.clo, "clo", (float, int, np.ndarray, list))
        if self.a_coefficient is not None:
            self.a_coefficient = convert_series_to_list(self.a_coefficient)
            validate_type(self.a_coefficient, "a_coefficient", (float, int))
        if self.e_coefficient is not None:
            self.e_coefficient = convert_series_to_list(self.e_coefficient)
            validate_type(self.e_coefficient, "e_coefficient", (float, int))
        if self.wme is not None:
            self.wme = convert_series_to_list(self.wme)
            validate_type(self.wme, "wme", (float, int, np.ndarray, list))
        if self.v_ankle is not None:
            self.v_ankle = convert_series_to_list(self.v_ankle)
            validate_type(self.v_ankle, "v_ankle", (float, int, np.ndarray, list))
        if self.t_running_mean is not None:
            self.t_running_mean = convert_series_to_list(self.t_running_mean)
            validate_type(
                self.t_running_mean,
                "t_running_mean",
                (float, int, np.ndarray, list),
            )
        if self.q is not None:
            self.q = convert_series_to_list(self.q)
            validate_type(self.q, "q", (float, int, np.ndarray, list))
        if not isinstance(self.round_output, bool):
            raise TypeError("round must be either True or False.")
        if not isinstance(self.limit_inputs, bool):
            raise TypeError("limit_inputs must be either True or False.")
        if not isinstance(self.airspeed_control, bool):
            raise TypeError("airspeed_control must be either True or False.")
        if not isinstance(self.with_solar_load, bool):
            raise TypeError("with_solar_load must be either True or False.")
        if self.tout is not None:
            self.tout = convert_series_to_list(self.tout)
            validate_type(self.tout, "tout", (float, int, np.ndarray, list))
        if self.p_atm is not None:
            self.p_atm = convert_series_to_list(self.p_atm)
            validate_type(self.p_atm, "p_atm", (float, int, np.ndarray, list))
        if self.age is not None:
            self.age = convert_series_to_list(self.age)
            validate_type(self.age, "age", (float, int, np.ndarray, list))
        if self.weight is not None:
            self.weight = convert_series_to_list(self.weight)
            validate_type(self.weight, "weight", (float, int, np.ndarray, list))
        if self.height is not None:
            self.height = convert_series_to_list(self.height)
            validate_type(self.height, "height", (float, int, np.ndarray, list))
        if self.sol_altitude is not None:
            self.sol_altitude = convert_series_to_list(self.sol_altitude)
            validate_type(
                self.sol_altitude,
                "sol_altitude",
                (float, int, np.ndarray, list),
            )
        if self.sharp is not None:
            self.sharp = convert_series_to_list(self.sharp)
            validate_type(self.sharp, "sharp", (float, int, np.ndarray, list))
        if self.sol_radiation_dir is not None:
            self.sol_radiation_dir = convert_series_to_list(self.sol_radiation_dir)
            validate_type(
                self.sol_radiation_dir,
                "sol_radiation_dir",
                (float, int, np.ndarray, list),
            )
        if self.sol_radiation_global is not None:
            self.sol_radiation_global = convert_series_to_list(
                self.sol_radiation_global,
            )
            validate_type(
                self.sol_radiation_global,
                "sol_radiation_global",
                (float, int, np.ndarray, list),
            )
        if self.sol_transmittance is not None:
            self.sol_transmittance = convert_series_to_list(self.sol_transmittance)
            validate_type(
                self.sol_transmittance,
                "sol_transmittance",
                (float, int, np.ndarray, list),
            )
        if self.f_svv is not None:
            self.f_svv = convert_series_to_list(self.f_svv)
            validate_type(self.f_svv, "f_svv", (float, int, np.ndarray, list))
        if self.f_bes is not None:
            self.f_bes = convert_series_to_list(self.f_bes)
            validate_type(self.f_bes, "f_bes", (float, int, np.ndarray, list))
        if self.asw is not None:
            self.asw = convert_series_to_list(self.asw)
            validate_type(self.asw, "asw", (float, int, np.ndarray, list))
        if self.floor_reflectance is not None:
            self.floor_reflectance = convert_series_to_list(self.floor_reflectance)
            validate_type(
                self.floor_reflectance,
                "floor_reflectance",
                (float, int, np.ndarray, list),
            )
        if self.vertical_tmp_grad is not None:
            self.vertical_tmp_grad = convert_series_to_list(self.vertical_tmp_grad)
            validate_type(
                self.vertical_tmp_grad,
                "vertical_tmp_grad",
                (float, int, np.ndarray, list),
            )
        if self.body_surface_area is not None:
            self.body_surface_area = convert_series_to_list(self.body_surface_area)
            validate_type(
                self.body_surface_area,
                "body_surface_area",
                (float, int, np.ndarray, list),
            )
        if self.max_sweating is not None:
            self.max_sweating = convert_series_to_list(self.max_sweating)
            validate_type(self.max_sweating, "max_sweating", (float, int))
        if self.max_skin_blood_flow is not None:
            self.max_skin_blood_flow = convert_series_to_list(self.max_skin_blood_flow)
            validate_type(self.max_skin_blood_flow, "max_skin_blood_flow", (float, int))
        if self.w_max is not None:
            self.w_max = convert_series_to_list(self.w_max)
            validate_type(self.w_max, "w_max", (float, int, np.ndarray, list))
        if self.wbgt is not None:
            self.wbgt = convert_series_to_list(self.wbgt)
            validate_type(self.wbgt, "wbgt", (float, int, np.ndarray, list))
        if self.work_intensity is not None:
            self.work_intensity = convert_series_to_list(self.work_intensity)
            _validate_str_values(
                "work_intensity",
                self.work_intensity,
                [i.value for i in WorkIntensity],
            )
        if self.thickness_quilt is not None:
            self.thickness_quilt = convert_series_to_list(self.thickness_quilt)
            validate_type(
                self.thickness_quilt,
                "thickness_quilt",
                (float, int, np.ndarray, list),
            )
        if self.vapor_pressure is not None:
            self.vapor_pressure = convert_series_to_list(self.vapor_pressure)
            validate_type(
                self.vapor_pressure,
                "vapor_pressure",
                (float, int, np.ndarray, list),
            )


@dataclass
class APMVInputs(BaseInputs):
    def __init__(
        self,
        tdb,
        tr,
        vr,
        rh,
        met,
        clo,
        a_coefficient,
        wme=0,
        units=Units.SI.value,
    ):
        # Initialize with only required fields, setting others to None
        super().__init__(
            tdb=tdb,
            tr=tr,
            vr=vr,
            rh=rh,
            met=met,
            clo=clo,
            a_coefficient=a_coefficient,
            wme=wme,
            units=units,
        )


@dataclass
class ASHRAEInputs(BaseInputs):
    def __init__(
        self,
        tdb,
        tr,
        t_running_mean,
        v,
        units,
    ):
        super().__init__(
            tdb=tdb,
            tr=tr,
            v=v,
            units=units,
            t_running_mean=t_running_mean,
        )


@dataclass
class ENInputs(BaseInputs):
    def __init__(
        self,
        tdb,
        tr,
        t_running_mean,
        v,
        units,
    ):
        super().__init__(
            tdb=tdb,
            tr=tr,
            v=v,
            units=units,
            t_running_mean=t_running_mean,
        )


@dataclass
class AnkleDraftInputs(BaseInputs):
    def __init__(
        self,
        tdb,
        tr,
        vr,
        rh,
        met,
        clo,
        v_ankle,
        units=Units.SI.value,
    ):
        # Initialize with only required fields, setting others to None
        super().__init__(
            tdb=tdb,
            tr=tr,
            vr=vr,
            rh=rh,
            met=met,
            clo=clo,
            v_ankle=v_ankle,
            units=units,
        )


@dataclass
class ATInputs(BaseInputs):
    def __init__(
        self,
        tdb,
        rh,
        v,
        q=None,
        round_output=True,
    ):
        # Initialize with only required fields, setting others to None
        super().__init__(
            tdb=tdb,
            vr=v,
            rh=rh,
            q=q,
            round_output=round_output,
        )


@dataclass
class ATHBInputs(BaseInputs):
    def __init__(
        self,
        tdb,
        tr,
        vr,
        rh,
        met,
        t_running_mean,
    ):
        super().__init__(
            tdb=tdb,
            tr=tr,
            vr=vr,
            rh=rh,
            met=met,
            t_running_mean=t_running_mean,
        )


@dataclass
class CloTOutInputs(BaseInputs):
    def __init__(
        self,
        tout,
        units: str = Units.SI.value,
    ):
        # Initialize with only required fields, setting others to None
        super().__init__(
            tout=tout,
            units=units,
        )


@dataclass
class CEInputs(BaseInputs):
    def __init__(
        self,
        tdb,
        tr,
        vr,
        rh,
        met,
        clo,
        wme,
        units,
    ):
        # Initialize with only required fields, setting others to None
        super().__init__(
            tdb=tdb,
            tr=tr,
            vr=vr,
            rh=rh,
            met=met,
            clo=clo,
            wme=wme,
            units=units,
        )


@dataclass
class DIInputs(BaseInputs):
    def __init__(
        self,
        tdb,
        rh,
    ):
        # Initialize with only required fields, setting others to None
        super().__init__(
            tdb=tdb,
            rh=rh,
        )


@dataclass
class EPMVInputs(BaseInputs):
    def __init__(
        self,
        tdb,
        tr,
        vr,
        rh,
        met,
        clo,
        e_coefficient,
        wme,
        units,
    ):
        # Initialize with only required fields, setting others to None
        super().__init__(
            tdb=tdb,
            tr=tr,
            vr=vr,
            rh=rh,
            met=met,
            clo=clo,
            e_coefficient=e_coefficient,
            wme=wme,
            units=units,
        )


@dataclass
class ESIInputs(BaseInputs):
    """Input class for the Environmental Stress Index (ESI) calculation.

    This class validates and processes inputs required for calculating the ESI, which
    evaluates heat stress based on temperature, humidity, and solar radiation.
    """

    def __init__(self, tdb, rh, sol_radiation_global, round_output=True):
        # Initialize with only required fields, setting others to None
        super().__init__(
            tdb=tdb,
            rh=rh,
            sol_radiation_global=sol_radiation_global,
            round_output=round_output,
        )

    def __post_init__(self):
        super().__post_init__()

        rh = np.asarray(self.rh, dtype=float)
        if np.any(rh < 0) or np.any(rh > 100):
            raise ValueError("Relative humidity must be between 0 and 100 %")

        sol_radiation_global = np.asarray(self.sol_radiation_global, dtype=float)
        if np.any(sol_radiation_global < 0):
            raise ValueError("Solar radiation must be greater than or equal to 0 W/m2")


class HIModels(Enum):
    rothfusz = "rothfusz"
    lu_romps = "lu-romps"


@dataclass
class HIInputs(BaseInputs):
    def __init__(
        self,
        tdb,
        rh,
        round_output,
        limit_inputs,
    ):
        # Initialize with only required fields, setting others to None
        super().__init__(
            tdb=tdb,
            rh=rh,
            round_output=round_output,
            limit_inputs=limit_inputs,
        )


class HumidexModels(Enum):
    rana = "rana"
    masterson = "masterson"


@dataclass
class HumidexInputs(BaseInputs):
    def __init__(
        self,
        tdb,
        rh,
        round_output,
    ):
        # Initialize with only required fields, setting others to None
        super().__init__(
            tdb=tdb,
            rh=rh,
            round_output=round_output,
        )


@dataclass
class NETInputs(BaseInputs):
    def __init__(
        self,
        tdb,
        rh,
        v,
        round_output=True,
    ):
        # Initialize with only required fields, setting others to None
        super().__init__(
            tdb=tdb,
            rh=rh,
            v=v,
            round_output=round_output,
        )


@dataclass
class PETSteadyInputs(BaseInputs):
    def __init__(
        self,
        tdb,
        tr,
        v,
        rh,
        met,
        clo,
        p_atm,
        position,
        age,
        sex,
        weight,
        height,
        wme,
    ):
        # Initialize with only required fields, setting others to None
        super().__init__(
            tdb=tdb,
            tr=tr,
            v=v,
            rh=rh,
            met=met,
            clo=clo,
            p_atm=p_atm,
            position=position,
            age=age,
            sex=sex,
            weight=weight,
            height=height,
            wme=wme,
        )


@dataclass
class PHSInputs(BaseInputs):
    def __init__(
        self,
        tdb,
        tr,
        v,
        rh,
        met,
        clo,
        round_output,
        wme,
        posture,
    ):
        # Initialize with only required fields, setting others to None
        super().__init__(
            tdb=tdb,
            tr=tr,
            v=v,
            rh=rh,
            met=met,
            clo=clo,
            round_output=round_output,
            wme=wme,
            posture=posture,
        )


@dataclass
class PMVInputs(BaseInputs):
    def __init__(
        self,
        tdb,
        tr,
        vr,
        rh,
        met,
        clo,
        wme=0,
        units=Units.SI.value,
        limit_inputs=True,
        airspeed_control=True,
    ):
        # Initialize with only required fields, setting others to None
        super().__init__(
            tdb=tdb,
            tr=tr,
            vr=vr,
            rh=rh,
            met=met,
            clo=clo,
            wme=wme,
            units=units,
            limit_inputs=limit_inputs,
            airspeed_control=airspeed_control,
        )


@dataclass
class PMVPPDInputs(BaseInputs):
    def __init__(
        self,
        tdb,
        tr,
        vr,
        rh,
        met,
        clo,
        wme=0,
        units=Units.SI.value,
        limit_inputs=True,
        airspeed_control=True,
    ):
        # Initialize with only required fields, setting others to None
        super().__init__(
            tdb=tdb,
            tr=tr,
            vr=vr,
            rh=rh,
            met=met,
            clo=clo,
            wme=wme,
            units=units,
            limit_inputs=limit_inputs,
            airspeed_control=airspeed_control,
        )


@dataclass
class SETInputs(BaseInputs):
    def __init__(
        self,
        tdb,
        tr,
        v,
        rh,
        met,
        clo,
        wme=0,
        body_surface_area=1.8258,
        p_atm=101325,
        position=Postures.standing.value,
        units=Units.SI.value,
        limit_inputs=True,
    ):
        # Initialize with only required fields, setting others to None
        super().__init__(
            tdb=tdb,
            tr=tr,
            v=v,
            rh=rh,
            met=met,
            clo=clo,
            wme=wme,
            body_surface_area=body_surface_area,
            p_atm=p_atm,
            position=position,
            units=units,
            limit_inputs=limit_inputs,
        )


@dataclass
class SolarGainInputs(BaseInputs):
    def __init__(
        self,
        sol_altitude,
        sharp,
        sol_radiation_dir,
        sol_transmittance,
        f_svv,
        f_bes,
        asw=0.7,
        posture=Postures.sitting.value,
        floor_reflectance=0.6,
    ):
        # Initialize with only required fields, setting others to None
        super().__init__(
            sol_altitude=sol_altitude,
            sharp=sharp,
            sol_radiation_dir=sol_radiation_dir,
            sol_transmittance=sol_transmittance,
            f_svv=f_svv,
            f_bes=f_bes,
            asw=asw,
            posture=posture,
            floor_reflectance=floor_reflectance,
        )


@dataclass
class GaggeTwoNodesInputs(BaseInputs):
    def __init__(
        self,
        tdb,
        tr,
        v,
        rh,
        met,
        clo,
        wme=0,
        body_surface_area=1.8258,
        p_atm=101325,
        position=Postures.standing.value,
        max_skin_blood_flow=90,
        round_output=True,
        max_sweating=500,
        w_max=None,
    ):
        # Initialize with only required fields, setting others to None
        super().__init__(
            tdb=tdb,
            tr=tr,
            v=v,
            rh=rh,
            met=met,
            clo=clo,
            wme=wme,
            body_surface_area=body_surface_area,
            p_atm=p_atm,
            position=position,
            max_skin_blood_flow=max_skin_blood_flow,
            round_output=round_output,
            max_sweating=max_sweating,
            w_max=w_max,
        )


@dataclass
class GaggeTwoNodesJiInputs(BaseInputs):
    def __init__(
        self,
        tdb,
        tr,
        v,
        met,
        clo,
        vapor_pressure,
        wme,
        body_surface_area,
        p_atm,
        position,
    ):
        # Initialize with only required fields, setting others to None
        super().__init__(
            tdb=tdb,
            tr=tr,
            v=v,
            met=met,
            clo=clo,
            vapor_pressure=vapor_pressure,
            wme=wme,
            body_surface_area=body_surface_area,
            p_atm=p_atm,
            position=position,
        )


@dataclass
class GaggeTwoNodesSleepInputs(BaseInputs):
    def __init__(
        self,
        tdb,
        tr,
        v,
        rh,
        clo,
        thickness_quilt,
        wme=0,
        p_atm=101325,
    ):
        # Initialise BaseInputs-supported fields
        super().__init__(
            tdb=tdb,
            tr=tr,
            v=v,
            rh=rh,
            clo=clo,
            wme=wme,
            p_atm=p_atm,
            thickness_quilt=thickness_quilt,
        )

    def __post_init__(self):
        super().__post_init__()

        if np.any(np.asarray(self.thickness_quilt, dtype=float) < 0):
            raise ValueError("thickness_quilt must be greater than or equal to 0 cm.")


@dataclass
class THIInputs(BaseInputs):
    def __init__(
        self,
        tdb,
        rh,
        round_output=True,
    ):
        # Initialize with only required fields, setting others to None
        super().__init__(
            tdb=tdb,
            rh=rh,
            round_output=round_output,
        )

    def __post_init__(self):
        super().__post_init__()

        rh = np.asarray(self.rh, dtype=float)
        if np.any(rh < 0) or np.any(rh > 100):
            raise ValueError("Relative humidity must be between 0 and 100 %")


@dataclass
class UseFansHeatwavesInputs(BaseInputs):
    def __init__(
        self,
        tdb,
        tr,
        v,
        rh,
        met,
        clo,
        wme=0,
        body_surface_area=1.8258,
        p_atm=101325,
        position=Postures.standing.value,
        max_skin_blood_flow=80,
        limit_inputs=True,
    ):
        # Initialize with only required fields, setting others to None
        super().__init__(
            tdb=tdb,
            tr=tr,
            v=v,
            rh=rh,
            met=met,
            clo=clo,
            wme=wme,
            body_surface_area=body_surface_area,
            p_atm=p_atm,
            position=position,
            max_skin_blood_flow=max_skin_blood_flow,
            limit_inputs=limit_inputs,
        )


@dataclass
class UTCIInputs(BaseInputs):
    def __init__(
        self,
        tdb,
        tr,
        v,
        rh,
        units=Units.SI.value,
        limit_inputs=True,
    ):
        # Initialize with only required fields, setting others to None
        super().__init__(
            tdb=tdb,
            tr=tr,
            v=v,
            rh=rh,
            units=units,
            limit_inputs=limit_inputs,
        )


@dataclass
class VerticalTGradPPDInputs(BaseInputs):
    def __init__(
        self,
        tdb,
        tr,
        vr,
        rh,
        met,
        clo,
        vertical_tmp_grad,
        units=Units.SI.value,
    ):
        # Initialize with only required fields, setting others to None
        super().__init__(
            tdb=tdb,
            tr=tr,
            vr=vr,
            rh=rh,
            met=met,
            clo=clo,
            vertical_tmp_grad=vertical_tmp_grad,
            units=units,
        )


@dataclass
class WBGTInputs(BaseInputs):
    def __init__(
        self,
        twb,
        tg,
        tdb=None,
        with_solar_load=False,
        round_output=True,
    ):
        # Initialize with only required fields, setting others to None
        super().__init__(
            twb=twb,
            tg=tg,
            tdb=tdb,
            with_solar_load=with_solar_load,
            round_output=round_output,
        )


@dataclass
class WCIInputs(BaseInputs):
    def __init__(
        self,
        tdb,
        v,
        round_output=True,
    ):
        # Initialize with only required fields, setting others to None
        super().__init__(
            tdb=tdb,
            v=v,
            round_output=round_output,
        )


@dataclass
class WCTInputs(BaseInputs):
    def __init__(
        self,
        tdb,
        v,
        round_output=True,
    ):
        # Initialize with only required fields, setting others to None
        super().__init__(
            tdb=tdb,
            v=v,
            round_output=round_output,
        )


@dataclass
class WorkCapacityHothapsInputs(BaseInputs):
    def __init__(
        self,
        wbgt,
        work_intensity,
    ):
        super().__init__(wbgt=wbgt, work_intensity=work_intensity)


@dataclass
class WorkCapacityStandardsInputs(BaseInputs):
    def __init__(
        self,
        wbgt,
        met,
    ):
        # Initialize with only required fields, setting others to None
        super().__init__(
            wbgt=wbgt,
            met=met,
        )

    def __post_init__(self):
        super().__post_init__()
        met = np.asarray(self.met, dtype=float)
        if np.any(met < 0) or np.any(met > 2500):
            raise ValueError("Metabolic rate out of plausible range")



import numpy as np

from pythermalcomfort.models import (
    set_tmp,
)
from tests.conftest import Urls, retrieve_reference_table, validate_result

tdb = []
tr = []
v = []
rh = []
met = []
clo = []
set_exp = []


def test_set_url(get_test_url, retrieve_data) -> None:
    """Test the SET model with various inputs."""
    reference_table = retrieve_reference_table(
        get_test_url,
        retrieve_data,
        Urls.SET.name,
    )
    tolerance = reference_table["tolerance"]
    for entry in reference_table["data"]:
        inputs = entry["inputs"]
        outputs = entry["outputs"]
        inputs["round_output"] = True
        inputs["limit_inputs"] = False
        result = set_tmp(**inputs)

        validate_result(result, outputs, tolerance)


def test_set_npnan() -> None:
    """Test that the function returns np.nan when inputs are np.nan."""
    np.testing.assert_equal(
        set_tmp(
            [41, 20, 20, 20, 20, 39],
            [20, 41, 20, 20, 20, 39],
            [0.1, 0.1, 2.1, 0.1, 0.1, 0.1],
            50,
            [1.1, 1.1, 1.1, 0.7, 1.1, 3.9],
            [0.5, 0.5, 0.5, 0.5, 2.1, 1.9],
        ).set,
        [np.nan, np.nan, np.nan, np.nan, np.nan, np.nan],
    )



from __future__ import annotations

import numpy as np

from pythermalcomfort.classes_input import WorkCapacityHothapsInputs, WorkIntensity
from pythermalcomfort.classes_return import WorkCapacity


def work_capacity_hothaps(
    wbgt: float | list[float],
    work_intensity: str = WorkIntensity.HEAVY.value,
) -> WorkCapacity:
    """Estimate work capacity due to heat based on Kjellstrom et al. [Kjellstrom2018]_.

    Estimates the amount of work that will be done at a given WBGT and
    intensity of work as a percent. 100% means work is unaffected by heat. 0%
    means no work is done. Note that in this version the functions do not reach
    0% as it is assumed that it is always possible to work in short bursts for
    10% of the time.

    Note that for this function "the empirical evidence is from studies in
    heavyly distinct locations, including a gold mine (Wyndham, 1969), 124 rice
    harvesters in West Bengal in India (Sahu et al., 2013), and six women
    observed in a climatic chamber (Nag and Nag, 1992)."
    (https://adaptecca.es/sites/default/files/documentos/2018_jrc_pesetaiii_impact_labour_productivity.pdf).
    The shape of the function is just an assumption, and the fit of the
    sigmoid to the data it is analysing is not especially good.

    Heavy intensity work is sometimes labelled as 400 W, moderate 300 W, light 200
    W, but this is only nominal.

    The correction citation is: Bröde P, Fiala D, Lemke B, Kjellstrom T.
    Estimated work ability in warm outdoor environments depends on the chosen
    heat stress assessment metric. International Journal of Biometeorology.
    2018 Mar;62(3):331 45.


    The relevant definitions of the functions can be found most clearly in:
    Orlov A, Sillmann J, Aunan K, Kjellstrom T, Aaheim A. Economic costs of
    heat-induced reductions in worker productivity due to global warming. Global
    Environmental Change [Internet]. 2020 Jul;63. Available from:
    https://doi.org/10.1016/j.gloenvcha.2020.102087

    For a comparison of different functions see Fig 1 of Day E, Fankhauser S, Kingsmill
    N, Costa H, Mavrogianni A. Upholding labour productivity under climate
    change: an assessment of adaptation options. Climate Policy. 2019
    Mar;19(3):367 85.

    Parameters
    ----------
    wbgt : float or list of floats
        Wet bulb globe temperature, [°C].
    work_intensity : str
        Which work intensity to use for the calculation, choice of "heavy",
        "moderate" or "light".

    Returns
    -------
    WorkCapacity
        A dataclass containing the work capacity. See
        :py:class:`~pythermalcomfort.classes_return.WorkCapacity` for more details. To access the
        `capacity` value, use the `capacity` attribute of the returned `WorkCapacity` instance, e.g.,
        `result.capacity`.


    """
    # validate inputs
    WorkCapacityHothapsInputs(wbgt=wbgt, work_intensity=work_intensity)

    # convert str to enum
    work_intensity = WorkIntensity(work_intensity.lower())
    wbgt = np.array(wbgt)

    params = {
        WorkIntensity.HEAVY: {"divisor": 30.94, "exponent": 16.64},
        WorkIntensity.MODERATE: {"divisor": 32.93, "exponent": 17.81},
        WorkIntensity.LIGHT: {"divisor": 34.64, "exponent": 22.72},
    }
    divisor = params[work_intensity]["divisor"]
    exponent = params[work_intensity]["exponent"]
    capacity = np.clip(100 * (0.1 + (0.9 / (1 + (wbgt / divisor) ** exponent))), 0, 100)

    return WorkCapacity(capacity=capacity)



import pytest

from pythermalcomfort.models import two_nodes_gagge_ji
from pythermalcomfort.utilities import body_surface_area, p_sat_torr


# Scenarios based on Table 4 from the paper by Ji et al. (2022)
@pytest.mark.parametrize(
    (
        "tdb",
        "tr",
        "v",
        "met",
        "clo",
        "rh",
        "weight",
        "height",
        "expected_t_core",
        "expected_t_skin",
        "duration",
    ),
    [
        # Scenario 1
        (
            36.5,  # tdb
            36.5,  # tr
            0.25,  # v
            0.95,  # met
            0.1,  # clo
            20,  # rh
            80.1,  # body_weight (kg)
            1.8,  # height (m)
            37.446339126254756,  # expected t_core[-1]
            34.596469729370725,  # expected t_skin[-1]
            120,  # length_time_simulation
        ),
        # Scenario 2
        (
            36.5,
            36.5,
            0.25,
            0.96,
            0.1,
            60,
            80.1,
            1.8,
            37.512644583008566,
            34.85861523864417,
            120,
        ),
        # Scenario 3
        (
            31,
            31,
            0.2,
            0.7,
            0.67,
            60,
            55.7,
            1.525,
            37.16672342773386,
            34.29506493710769,
            150,
        ),
        # Scenario 4
        (
            27,
            27,
            0.2,
            0.7,
            0.67,
            60,
            55.7,
            1.525,
            36.69341582589469,
            33.64236786746656,
            150,
        ),
    ],
)
def test_two_nodes_gagge_ji_examples(
    tdb: float,
    tr: float,
    v: float,
    met: float,
    clo: float,
    rh: float,
    weight: float,
    height: float,
    expected_t_core: float,
    expected_t_skin: float,
    duration: int,
) -> None:
    """Test the two_nodes_gagge_ji function with predefined scenarios."""
    # compute vapor pressure from relative humidity and saturation vapor pressure
    vapor_pressure = rh * p_sat_torr(tdb=tdb) / 100

    # compute body surface area
    bsa = body_surface_area(weight=weight, height=height)

    result = two_nodes_gagge_ji(
        tdb=tdb,
        tr=tr,
        v=v,
        met=met,
        clo=clo,
        vapor_pressure=vapor_pressure,
        wme=0,
        body_surface_area=bsa,
        p_atm=101325,
        position="sitting",
        acclimatized=True,
        body_weight=weight,
        length_time_simulation=duration,
    )

    t_core_last = result["t_core"][-1]
    t_skin_last = result["t_skin"][-1]

    assert t_core_last == pytest.approx(expected_t_core, rel=1e-3)
    assert t_skin_last == pytest.approx(expected_t_skin, rel=1e-3)


def test_two_nodes_gagge_ji_list_tdb() -> None:
    """Test that the function can handle a list of tdb values."""
    tdb_list = [36.5, 36.5]
    tr = 36.5
    v = 0.25
    met = 0.95
    clo = 0.1
    rh = 20
    weight = 80.1
    height = 1.8
    sim_time = 120

    vapor_pressure = rh * p_sat_torr(tdb=36.5) / 100
    bsa = body_surface_area(weight=weight, height=height)

    result = two_nodes_gagge_ji(
        tdb=tdb_list,
        tr=tr,
        v=v,
        met=met,
        clo=clo,
        vapor_pressure=vapor_pressure,
        wme=0,
        body_surface_area=bsa,
        p_atm=101325,
        position="sitting",
        acclimatized=True,
        body_weight=weight,
        length_time_simulation=sim_time,
    )

    # expect two simulation results (one per tdb entry)
    t_core_list = result.t_core
    t_skin_list = result.t_skin

    assert isinstance(t_core_list, list)
    assert len(t_core_list) == 2
    assert isinstance(t_skin_list, list)
    assert len(t_skin_list) == 2

    # both simulations should yield the same final values as scenario 1
    expected_t_core = 37.446
    expected_t_skin = 34.596

    for core_arr, skin_arr in zip(t_core_list, t_skin_list, strict=False):
        assert core_arr[-1] == pytest.approx(expected_t_core, rel=1e-3)
        assert skin_arr[-1] == pytest.approx(expected_t_skin, rel=1e-3)


def test_invalid_position_raises_value_error() -> None:
    """Test that an invalid position raises a ValueError."""
    with pytest.raises(ValueError):
        two_nodes_gagge_ji(
            tdb=36.5,
            tr=36.5,
            v=0.25,
            met=0.95,
            clo=0.1,
            vapor_pressure=20 * p_sat_torr(tdb=36.5) / 100,
            wme=0,
            body_surface_area=body_surface_area(weight=80.1, height=1.8),
            p_atm=101325,
            position="lying",  # invalid posture
            acclimatized=True,
            body_weight=80.1,
            length_time_simulation=120,
        )


def test_unexpected_kwarg_raises_type_error() -> None:
    """Test that an unexpected keyword argument raises a TypeError."""
    with pytest.raises(TypeError) as excinfo:
        two_nodes_gagge_ji(
            tdb=36.5,
            tr=36.5,
            v=0.25,
            met=0.95,
            clo=0.1,
            vapor_pressure=20 * p_sat_torr(tdb=36.5) / 100,
            wme=0,
            body_surface_area=body_surface_area(weight=80.1, height=1.8),
            p_atm=101325,
            position="sitting",
            acclimatized=True,
            body_weight=80.1,
            length_time_simulation=120,
            foo="bar",  # unexpected argument
        )
    assert "Unexpected keyword arguments: ['foo']" in str(excinfo.value)


def test_non_numeric_tdb_raises_type_error() -> None:
    """Test that a non-numeric tdb raises a TypeError."""
    with pytest.raises(TypeError):
        two_nodes_gagge_ji(
            tdb="hot",  # invalid type
            tr=36.5,
            v=0.25,
            met=0.95,
            clo=0.1,
            vapor_pressure=20 * p_sat_torr(tdb=36.5) / 100,
            wme=0,
            body_surface_area=body_surface_area(weight=80.1, height=1.8),
            p_atm=101325,
            position="sitting",
            acclimatized=True,
            body_weight=80.1,
            length_time_simulation=120,
        )



from __future__ import annotations

import math

import numpy as np

from pythermalcomfort.classes_input import SolarGainInputs
from pythermalcomfort.classes_return import SolarGain
from pythermalcomfort.utilities import Postures, transpose_sharp_altitude


def solar_gain(
    sol_altitude: float | list[float],
    sharp: float | list[float],
    sol_radiation_dir: float | list[float],
    sol_transmittance: float | list[float],
    f_svv: float | list[float],
    f_bes: float | list[float],
    asw: float | list[float] = 0.7,
    posture: str = Postures.sitting.value,
    floor_reflectance: float | list[float] = 0.6,
    round_output: bool = True,
) -> SolarGain:
    """Calculate the solar gain to the human body using the Effective Radiant
    Field (ERF) [55ASHRAE2023]_. The ERF is a measure of the net energy flux to or from
    the human body. ERF is expressed in W over human body surface area [W/m2].
    In addition, it calculates the delta mean radiant temperature. Which is the
    amount by which the mean radiant temperature of the space should be
    increased if no solar radiation is present.

    Parameters
    ----------
    sol_altitude : float or list of floats
        Solar altitude, degrees from horizontal [deg]. Ranges between 0 and 90.
    sharp : float or list of floats
        Solar horizontal angle relative to the front of the person (SHARP) [deg].
        Ranges between 0 and 180 and is symmetrical on either side. Zero (0) degrees
        represents direct-beam radiation from the front, 90 degrees represents
        direct-beam radiation from the side, and 180 degrees represents direct-beam
        radiation from the back. SHARP is the angle between the sun and the person
        only. Orientation relative to compass or to room is not included in SHARP.
    sol_radiation_dir : float or list of floats
        Direct-beam solar radiation, [W/m2]. Ranges between 200 and 1000. See Table
        C2-3 of ASHRAE 55 2020 [55ASHRAE2023]_.
    sol_transmittance : float or list of floats
        Total solar transmittance, ranges from 0 to 1. The total solar
        transmittance of window systems, including glazing unit, blinds, and other
        façade treatments, shall be determined using one of the following methods:
        i) Provided by manufacturer or from the National Fenestration Rating
        Council approved Lawrence Berkeley National Lab International Glazing
        Database.
        ii) Glazing unit plus venetian blinds or other complex or unique shades
        shall be calculated using National Fenestration Rating Council approved
        software or Lawrence Berkeley National Lab Complex Glazing Database.
    f_svv : float or list of floats
        Fraction of sky-vault view fraction exposed to body, ranges from 0 to 1.
        It can be calculated using the function
        :py:meth:`pythermalcomfort.utilities.f_svv`.
    f_bes : float or list of floats
        Fraction of the possible body surface exposed to sun, ranges from 0 to 1.
        See Table C2-2 and equation C-7 ASHRAE 55 2020 [55ASHRAE2023]_.
    asw : float or list of floats, optional
        The average short-wave absorptivity of the occupant. It will range widely,
        depending on the color of the skin of the occupant as well as the color and
        amount of clothing covering the body. Defaults to 0.7.

        .. note::
            Short-wave absorptivity typically ranges from 0.57 to 0.84, depending
            on skin and clothing color. More information is available in Blum (1945).

    posture : str, optional
        Default 'sitting' list of available options 'standing', 'supine' or 'sitting'.
    floor_reflectance : float or list of floats, optional
        Floor reflectance. It is assumed to be constant and equal to 0.6. Defaults to 0.6.
    round_output : bool, optional
        If True, rounds output value. If False, it does not round it. Defaults to True.

    Returns
    -------
    SolarGain
        A dataclass containing the solar gain to the human body and delta mean radiant temperature.
        See :py:class:`~pythermalcomfort.classes_return.SolarGain` for more details.
        To access the `erf` and `delta_mrt` values, use the corresponding attributes of the returned `SolarGain` instance, e.g., `result.erf`.

    Examples
    --------
    .. code-block:: python

        from pythermalcomfort.models import solar_gain

        result = solar_gain(
            sol_altitude=0,
            sharp=120,
            sol_radiation_dir=800,
            sol_transmittance=0.5,
            f_svv=0.5,
            f_bes=0.5,
            asw=0.7,
            posture="sitting",
        )
        print(result.erf)  # 42.9
        print(result.delta_mrt)  # 10.3

    """
    # Validate inputs using the SolarGainInputs class
    SolarGainInputs(
        sol_altitude=sol_altitude,
        sharp=sharp,
        sol_radiation_dir=sol_radiation_dir,
        sol_transmittance=sol_transmittance,
        f_svv=f_svv,
        f_bes=f_bes,
        asw=asw,
        floor_reflectance=floor_reflectance,
    )

    sol_altitude = np.array(sol_altitude)
    sharp = np.array(sharp)
    sol_radiation_dir = np.array(sol_radiation_dir)
    sol_transmittance = np.array(sol_transmittance)
    f_svv = np.array(f_svv)
    f_bes = np.array(f_bes)
    asw = np.array(asw)
    floor_reflectance = np.array(floor_reflectance)

    posture = posture.lower()
    if posture not in [
        Postures.standing.value,
        Postures.supine.value,
        Postures.sitting.value,
    ]:
        error_msg_posture = (
            "Posture has to be either 'standing', 'supine' or 'sitting'."
        )
        raise ValueError(error_msg_posture)

    erf, d_mrt = _solar_gain_vectorised(
        sol_altitude=sol_altitude,
        sharp=sharp,
        sol_radiation_dir=sol_radiation_dir,
        sol_transmittance=sol_transmittance,
        f_svv=f_svv,
        f_bes=f_bes,
        asw=asw,
        floor_reflectance=floor_reflectance,
        posture=posture,
    )

    if round_output:
        erf = np.round(erf, 1)
        d_mrt = np.round(d_mrt, 1)

    return SolarGain(erf=erf, delta_mrt=d_mrt)


@np.vectorize
def _solar_gain_vectorised(
    sol_altitude,
    sharp,
    sol_radiation_dir,
    sol_transmittance,
    f_svv,
    f_bes,
    asw,
    posture,
    floor_reflectance,
):
    def find_span(arr, x):
        for i in range(len(arr)):
            if arr[i + 1] >= x >= arr[i]:
                return i
        return -1

    deg_to_rad = 0.0174532925
    hr = 6
    i_diff = 0.2 * sol_radiation_dir

    # fp is the projected area factor
    fp_table = [
        [0.35, 0.35, 0.314, 0.258, 0.206, 0.144, 0.082],
        [0.342, 0.342, 0.31, 0.252, 0.2, 0.14, 0.082],
        [0.33, 0.33, 0.3, 0.244, 0.19, 0.132, 0.082],
        [0.31, 0.31, 0.275, 0.228, 0.175, 0.124, 0.082],
        [0.283, 0.283, 0.251, 0.208, 0.16, 0.114, 0.082],
        [0.252, 0.252, 0.228, 0.188, 0.15, 0.108, 0.082],
        [0.23, 0.23, 0.214, 0.18, 0.148, 0.108, 0.082],
        [0.242, 0.242, 0.222, 0.18, 0.153, 0.112, 0.082],
        [0.274, 0.274, 0.245, 0.203, 0.165, 0.116, 0.082],
        [0.304, 0.304, 0.27, 0.22, 0.174, 0.121, 0.082],
        [0.328, 0.328, 0.29, 0.234, 0.183, 0.125, 0.082],
        [0.344, 0.344, 0.304, 0.244, 0.19, 0.128, 0.082],
        [0.347, 0.347, 0.308, 0.246, 0.191, 0.128, 0.082],
    ]
    if posture == Postures.sitting.value:
        fp_table = [
            [0.29, 0.324, 0.305, 0.303, 0.262, 0.224, 0.177],
            [0.292, 0.328, 0.294, 0.288, 0.268, 0.227, 0.177],
            [0.288, 0.332, 0.298, 0.29, 0.264, 0.222, 0.177],
            [0.274, 0.326, 0.294, 0.289, 0.252, 0.214, 0.177],
            [0.254, 0.308, 0.28, 0.276, 0.241, 0.202, 0.177],
            [0.23, 0.282, 0.262, 0.26, 0.233, 0.193, 0.177],
            [0.216, 0.26, 0.248, 0.244, 0.22, 0.186, 0.177],
            [0.234, 0.258, 0.236, 0.227, 0.208, 0.18, 0.177],
            [0.262, 0.26, 0.224, 0.208, 0.196, 0.176, 0.177],
            [0.28, 0.26, 0.21, 0.192, 0.184, 0.17, 0.177],
            [0.298, 0.256, 0.194, 0.174, 0.168, 0.168, 0.177],
            [0.306, 0.25, 0.18, 0.156, 0.156, 0.166, 0.177],
            [0.3, 0.24, 0.168, 0.152, 0.152, 0.164, 0.177],
        ]

    if posture == Postures.supine.value:
        sharp, sol_altitude = transpose_sharp_altitude(sharp, sol_altitude)

    alt_range = [0, 15, 30, 45, 60, 75, 90]
    az_range = [0, 15, 30, 45, 60, 75, 90, 105, 120, 135, 150, 165, 180]
    alt_i = find_span(alt_range, sol_altitude)
    az_i = find_span(az_range, sharp)
    fp11 = fp_table[az_i][alt_i]
    fp12 = fp_table[az_i][alt_i + 1]
    fp21 = fp_table[az_i + 1][alt_i]
    fp22 = fp_table[az_i + 1][alt_i + 1]
    az1 = az_range[az_i]
    az2 = az_range[az_i + 1]
    alt1 = alt_range[alt_i]
    alt2 = alt_range[alt_i + 1]
    fp = fp11 * (az2 - sharp) * (alt2 - sol_altitude)
    fp += fp21 * (sharp - az1) * (alt2 - sol_altitude)
    fp += fp12 * (az2 - sharp) * (sol_altitude - alt1)
    fp += fp22 * (sharp - az1) * (sol_altitude - alt1)
    fp /= (az2 - az1) * (alt2 - alt1)

    f_eff = 0.725  # fraction of the body surface exposed to environmental radiation
    if posture == Postures.sitting.value:
        f_eff = 0.696

    sw_abs = asw
    lw_abs = 0.95

    e_diff = f_eff * f_svv * 0.5 * sol_transmittance * i_diff
    e_direct = f_eff * fp * sol_transmittance * f_bes * sol_radiation_dir
    e_reflected = (
        f_eff
        * f_svv
        * 0.5
        * sol_transmittance
        * (sol_radiation_dir * math.sin(sol_altitude * deg_to_rad) + i_diff)
        * floor_reflectance
    )

    e_solar = e_diff + e_direct + e_reflected
    erf = e_solar * (sw_abs / lw_abs)
    d_mrt = erf / (hr * f_eff)

    return erf, d_mrt



"""Test suite for Temperature-Humidity Index (THI) model.

Tests cover scalar and list inputs, rounding behaviour, and input validation.
"""

import numpy as np
import pytest

from pythermalcomfort.classes_return import THI
from pythermalcomfort.models import thi
from tests.conftest import is_equal


@pytest.mark.parametrize(
    ("tdb", "rh", "expected"),
    [
        (30.0, 70.0, 81.4),
        (20.0, 50.0, 65.2),
    ],
)
def test_scalar_rounding_default(tdb: float, rh: float, expected: float) -> None:
    """Test that the function calculates THI correctly with default rounding."""
    result = thi(tdb, rh)
    assert isinstance(result, THI)
    assert isinstance(result.thi, float)
    assert is_equal(result.thi, expected)


def test_scalar_no_rounding() -> None:
    """Test that the function calculates THI correctly without rounding."""
    tdb, rh = 30.0, 70.0
    expected = 1.8 * tdb + 32 - 0.55 * (1 - 0.01 * rh) * (1.8 * tdb - 26)
    result = thi(tdb, rh, round_output=False)
    assert is_equal(result.thi, expected)


def test_list_input() -> None:
    """Test that the function calculates THI correctly for lists of tdb and rh values."""
    tdb_list = [30, 20]
    rh_list = [70, 50]
    result = thi(tdb_list, rh_list)
    expected = [81.4, 65.2]

    assert isinstance(result.thi, np.ndarray)
    assert is_equal(result.thi, expected)


@pytest.mark.parametrize(
    ("tdb", "rh", "expected_error"),
    [
        (25.0, -5.0, ValueError),
        (25.0, 150.0, ValueError),  # humidity should be between 0 and 100
        ("hot", "humid", TypeError),  # invalid types
    ],
)
def test_invalid_inputs_raise_specific(
    tdb: float,
    rh: float,
    expected_error: Exception,
) -> None:
    """Test that the function raises specific errors for invalid inputs."""
    with pytest.raises(expected_error):
        thi(tdb, rh)
